
/*******************************************************************************
*
* File check8.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Simple checks of the deflation preconditioner
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "dfl.h"
#include "global.h"


static void random_basis(int Ns)
{
   int i;
   spinor **ws;

   ws=reserve_ws(Ns);

   for (i=0;i<Ns;i++)
      random_s(VOLUME,ws[i],1.0f);
   
   dfl_subspace(ws);

   release_ws();
}


static void random_mode(int k)
{
   int nb,isw,Ns,nv;
   complex_dble **wvd;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*Ns;

   wvd=reserve_wvd(1);

   random_vd(nv,wvd[0],1.0);
   vnormalize_dble(nv,1,wvd[0]);
   dfl_vd2sd(wvd[0],k);

   release_wvd();
}


int main(int argc,char *argv[])
{
   int my_rank,bs[4],nb,isw,Ns,nv;
   int vnkv,vnmx,vdnmx,status;
   int iwd0,iwd1,iwd2,iwd3,iws0,iws1;
   double vres,vdres,dev0,dev1,dev2,dev3;
   complex zs;
   complex_dble **wvd,z,w0,w1;
   spinor **ws;
   spinor_dble **wsd;
   dfl_parms_t dfl;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check8.log","w",stdout);
      fin=freopen("check8.in","r",stdin);
      
      printf("\n");
      printf("Simple checks of the deflation preconditioner\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      read_line("vnkv","%d",&vnkv);
      read_line("vnmx","%d",&vnmx);
      read_line("vdnmx","%d",&vdnmx);
      read_line("vres","%lf",&vres);      
      read_line("vdres","%lf",&vdres);      
      fclose(fin);

      if (Ns<2)
         Ns=2;
      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n",Ns);
      printf("vnkv = %d\n",vnkv);
      printf("vnmx = %d\n",vnmx);
      printf("vdnmx = %d\n",vdnmx);
      printf("vres = %.1e\n",vres);
      printf("vdres = %.1e\n\n",vdres);            

      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   start_ranlux(0,123456);   
   geometry();

   alloc_ud();
   alloc_u();
   alloc_swd();
   alloc_sw();
   alloc_s(Ns+4);
   alloc_sd(5);

   set_lat_parms(0.0,0.12,1.0);
   dfl=set_dfl_parms(bs,Ns,4,1,1,1,vnkv,vnmx,vdnmx,0.12,1.0e-3,vres,vdres,0,3);

   error((vnkv!=dfl.vnkv)||(vnmx!=dfl.vnmx)||(vdnmx!=dfl.vdnmx)||
         (vres!=dfl.vres)||(vdres!=dfl.vdres),1,"main [check8.c]",
         "Parameter vnkv,vnmx,vdnmx,vres or vdres is incorrectly set");
   
   random_ud();
   assign_ud2u();
   sw_term();
   assign_swd2sw();
   random_basis(Ns);
   
   wsd=reserve_wsd(4);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;
   iwd2=(wsd[2]-psd[0][0])/NSPIN;
   iwd3=(wsd[3]-psd[0][0])/NSPIN;

   ws=reserve_ws(4);
   iws0=(ws[0]-ps[0][0])/NSPIN;
   iws1=(ws[1]-ps[0][0])/NSPIN;

   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*Ns;
   
   zs.re=-1.0f;
   zs.im=0.0f;
   z.re=-1.0;
   z.im=0.0;
   
   random_mode(iwd0);
   dfl_Pproj_dble(iwd0,iwd1);
   mulc_spinor_add_dble(VOLUME,wsd[0],wsd[1],z);
   dev0=norm_square_dble(VOLUME,1,wsd[0])/
      norm_square_dble(VOLUME,1,wsd[1]);
   dev0=sqrt(dev0);

   random_sd(VOLUME,wsd[0],1.0);
   dfl_Pproj_dble(iwd0,iwd1);
   dfl_Pproj_dble(iwd1,iwd2);
   mulc_spinor_add_dble(VOLUME,wsd[1],wsd[2],z);
   dev1=norm_square_dble(VOLUME,1,wsd[1])/
      norm_square_dble(VOLUME,1,wsd[2]);
   dev1=sqrt(dev1);
   
   random_sd(VOLUME,wsd[0],1.0);
   random_sd(VOLUME,wsd[1],1.0);
   assign_sd2sd(VOLUME,wsd[0],wsd[2]);
   assign_sd2sd(VOLUME,wsd[1],wsd[3]);
   dfl_Pproj_dble(iwd0,iwd0);
   dfl_Pproj_dble(iwd1,iwd1);
   w0=spinor_prod_dble(VOLUME,1,wsd[3],wsd[0]);
   w1=spinor_prod_dble(VOLUME,1,wsd[1],wsd[2]);
   w0.re-=w1.re;
   w0.im-=w1.im;
   dev2=(w0.re*w0.re+w0.im*w0.im)/
      (norm_square_dble(VOLUME,1,wsd[3])*norm_square_dble(VOLUME,1,wsd[0]));
   dev2=sqrt(dev2);

   message("Check of dfl_Pproj_dble():\n");
   message("Action in subspace: %.1e\n",dev0);
   message("Projector property: %.1e\n",dev1);
   message("Hermiticity check : %.1e\n\n",dev2);

   random_sd(VOLUME,wsd[0],1.0);
   assign_sd2sd(VOLUME,wsd[0],wsd[3]);
   dfl_Pproj_dble(iwd0,iwd1);
   dfl_Qproj_dble(iwd0,iwd2);
   mulc_spinor_add_dble(VOLUME,wsd[3],wsd[1],z);
   mulc_spinor_add_dble(VOLUME,wsd[3],wsd[2],z);
   dev0=norm_square_dble(VOLUME,1,wsd[3])/
      norm_square_dble(VOLUME,1,wsd[0]);
   dev0=sqrt(dev0);

   random_sd(VOLUME,wsd[0],1.0);
   dfl_Pproj_dble(iwd0,iwd1);
   dfl_Qproj_dble(iwd1,iwd2);
   dev1=norm_square_dble(VOLUME,1,wsd[2])/
      norm_square_dble(VOLUME,1,wsd[0]);
   dev1=sqrt(dev1);
   
   message("Check of dfl_Qproj_dble():\n");
   message("Q+P-1: %.1e\n",dev0);
   message("Q*P  : %.1e\n\n",dev1);

   wvd=reserve_wvd(3);
   random_vd(nv,wvd[0],1.0);
   assign_vd2vd(nv,wvd[0],wvd[1]);
   dfl_solve_little(vdnmx,vdres,wvd[0],&status);
   Znohat_dble(wvd[0],wvd[2]);
   mulc_vadd_dble(nv,wvd[2],wvd[1],z);
   dev0=vnorm_square_dble(nv,1,wvd[2])/vnorm_square_dble(nv,1,wvd[1]);
   dev0=sqrt(dev0);
   release_wvd();

   message("Check of dfl_solve_little():\n");
   message("Status = %d\n",status);
   message("Relative residue = %.1e (should be less than %.1e)\n\n",dev0,vdres);
   
   random_sd(VOLUME,wsd[0],1.0);
   random_sd(VOLUME,wsd[1],1.0);
   assign_sd2sd(VOLUME,wsd[0],wsd[2]);
   dfl_Lpcnd_dble(iwd0,iwd1,&status);
   dfl_Pproj_dble(iwd0,iwd3);
   dev0=norm_square_dble(VOLUME,1,wsd[3])/
      norm_square_dble(VOLUME,1,wsd[0]);
   dev0=sqrt(dev0);

   mulc_spinor_add_dble(VOLUME,wsd[2],wsd[0],z);
   Qnohat_dble(iwd1,iwd3);
   mulg5_dble(VOLUME,wsd[3]);   
   mulc_spinor_add_dble(VOLUME,wsd[3],wsd[2],z);
   dev1=norm_square_dble(VOLUME,1,wsd[3])/
      norm_square_dble(VOLUME,1,wsd[2]);
   dev1=sqrt(dev1);

   random_sd(VOLUME,wsd[0],1.0);
   dfl_Qproj_dble(iwd0,iwd0);
   assign_sd2sd(VOLUME,wsd[0],wsd[2]);   
   dfl_Lpcnd_dble(iwd0,iwd1,&status);
   mulc_spinor_add_dble(VOLUME,wsd[0],wsd[2],z);
   dev2=norm_square_dble(VOLUME,1,wsd[0])/
      norm_square_dble(VOLUME,1,wsd[2]);
   dev2=sqrt(dev2);   

   random_mode(iwd0);
   Qnohat_dble(iwd0,iwd1);
   mulg5_dble(VOLUME,wsd[1]);
   assign_sd2sd(VOLUME,wsd[1],wsd[3]);
   dfl_Lpcnd_dble(iwd1,iwd2,&status);
   dev3=norm_square_dble(VOLUME,1,wsd[1])/
      norm_square_dble(VOLUME,1,wsd[3]);
   dev3=sqrt(dev3);     
   
   message("Check of dfl_Lpcnd_dble():\n");
   message("Status = %d\n",status);
   message("P*PL   : %.1e\n",dev0);   
   message("D*phi  : %.1e\n",dev1);   
   message("PL*Q   : %.1e\n",dev2);
   message("PL*D*P : %.1e\n\n",dev3);         

   random_sd(VOLUME,wsd[0],1.0);
   Qnohat_dble(iwd0,iwd1);
   mulg5_dble(VOLUME,wsd[1]);
   assign_sd2sd(VOLUME,wsd[1],wsd[2]);      
   dfl_Lpcnd_dble(iwd2,iwd3,&status);      
   dfl_RLpcnd_dble(iwd0,iwd1,&status);
   mulc_spinor_add_dble(VOLUME,wsd[2],wsd[1],z);
   dev0=norm_square_dble(VOLUME,1,wsd[2])/
      norm_square_dble(VOLUME,1,wsd[1]);
   dev0=sqrt(dev0);
   
   Qnohat_dble(iwd0,iwd2);
   mulg5_dble(VOLUME,wsd[2]);
   mulc_spinor_add_dble(VOLUME,wsd[2],wsd[1],z);
   dev1=norm_square_dble(VOLUME,1,wsd[2])/
      norm_square_dble(VOLUME,1,wsd[1]);
   dev1=sqrt(dev1);
   
   message("Check of dfl_RLpcnd_dble():\n");
   message("Status = %d\n",status);
   message("PL       : %.1e\n",dev0);
   message("D*PR-PL*D: %.1e\n\n",dev1);      

   random_s(VOLUME,ws[0],1.0f);
   assign_s2sd(VOLUME,ws[0],wsd[0]);
   Qnohat(iws0,iws1);
   mulg5(VOLUME,ws[1]);
   Qnohat_dble(iwd0,iwd1);
   mulg5_dble(VOLUME,wsd[1]);   
   dfl_RLpcnd_dble(iwd0,iwd1,&status);
   dfl_RLpcnd(iws0,iws1,&status);   
   assign_sd2s(VOLUME,wsd[0],ws[2]);
   assign_sd2s(VOLUME,wsd[1],ws[3]);    

   mulc_spinor_add(VOLUME,ws[2],ws[0],zs);
   dev0=(double)(norm_square(VOLUME,1,ws[2])/
                 norm_square(VOLUME,1,ws[0]));
   dev0=sqrt(dev0);   
   
   mulc_spinor_add(VOLUME,ws[3],ws[1],zs);
   dev1=(double)(norm_square(VOLUME,1,ws[3])/
                 norm_square(VOLUME,1,ws[1]));
   dev1=sqrt(dev1);

   message("Check of dfl_RLpcnd():\n");
   message("Status = %d\n",status);
   message("PR: %.1e\n",dev0);
   message("PL: %.1e\n\n",dev1); 

   error_chk();   
   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();   
   exit(0);
}
