
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the linalg_dble routines
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "global.h"

static complex_dble v[25];
static complex_dble w[25];
static spinor_dble *ppk[10];


int main(int argc,char *argv[])
{
   int my_rank,count,i,j,n;
   int vol,icom;
   double r,wt1,wt2,wdt;
   complex_dble z,mz;
   spinor_dble *pk,*pl;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
      
      printf("\n");
      printf("Timing of the linalg_dble routines\n");
      printf("----------------------------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   vol=VOLUME;
   icom=1;
   
   start_ranlux(0,12345);
   geometry();
   alloc_sd(10);

   for (i=0;i<10;i++)
      random_sd(vol,psd[i][0],1.0f);

   n=(int)(1.0e7/(double)(vol));
   if (n<1)
      n=1;

   pk=psd[0][0];
   pl=psd[1][0];

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
      z=spinor_prod_dble(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function spinor_prod_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor_dble)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      r=spinor_prod_re_dble(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function spinor_prod_re_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(48.0/wdt),sizeof(spinor_dble)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      r=norm_square_dble(vol,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function norm_square_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(48.0/wdt),sizeof(spinor_dble)/3);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      normalize_dble(vol,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function normalize_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(72.0/wdt),sizeof(spinor_dble)/3);
   }
   
   mz.re=-z.re;
   mz.im=-z.im;

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      mulc_spinor_add_dble(vol,pk,pl,z);
      mulc_spinor_add_dble(vol,pk,pl,mz);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(2*n*vol);

   if (my_rank==0)
   {   
      printf("Function mulc_spinor_add_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor_dble)/3);
   }
   
   random_sd(vol,psd[0][0],1.0f);
   random_sd(vol,psd[1][0],1.0f);
   normalize_dble(vol,icom,pk);
   normalize_dble(vol,icom,pl);
   project_dble(vol,icom,pl,pk);
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
      project_dble(vol,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*vol);

   if (my_rank==0)
   {   
      printf("Function project_dble:\n");
      printf("Time per lattice point: %4.3f micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(192.0/wdt),sizeof(spinor_dble)/3);
   }
   
   for (i=0;i<5;i++)
   {
      random_sd(vol,psd[i][0],1.0f);
      ppk[i]=psd[i][0];

      for (j=0;j<i;j++)
         project_dble(vol,icom,ppk[i],ppk[j]);

      normalize_dble(vol,icom,ppk[i]);
   }

   for (i=5;i<10;i++)
   {
      random_sd(vol,psd[i][0],1.0f);
      ppk[i]=psd[i][0];

      for (j=5;j<i;j++)
         project_dble(vol,icom,ppk[i],ppk[j]);

      normalize_dble(vol,icom,ppk[i]);
   }   
      
   for (i=0;i<5;i++)
   {
      for (j=0;j<5;j++)
      {
         z=spinor_prod_dble(vol,icom,ppk[i],ppk[j+5]);

         v[5*i+j].re= z.re;
         v[5*i+j].im= z.im;
         w[5*j+i].re= z.re;
         w[5*j+i].im=-z.im;         
      }
   }

   n/=25;
   if (n<1)
      n=1;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      if ((count&0x7)==0x0)
      {
         for (i=0;i<5;i++)
            assign_sd2sd(vol,ppk[i+5],ppk[i]);
      }
      
      rotate_dble(vol,5,ppk,v);
      rotate_dble(vol,5,ppk,w);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   error_chk();   
   wdt=1.0e6*(wt2-wt1)/(double)(50*n*vol);

   if (my_rank==0)
   {   
      printf("Function rotate_dble (n=5 fields):\n");
      printf("Time per lattice point: %4.3f*n^2 micro sec\n",wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(96.0/wdt),sizeof(spinor_dble)/3);
      fclose(flog);
   }

   MPI_Finalize();   
   exit(0);
}
