
/*******************************************************************************
*
* File Zhat.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Application of the single-precision little Dirac operator
*
* The externally accessible functions are
*
*   void Zhat(complex *v,complex *w)
*     Applies the even-odd preconditioned little Dirac operator to the
*     even components of the global single-precision vector field v[] and
*     assigns the result to the field w[]
*
*   void Zeeinv(complex *v,complex *w)
*     Applies the inverse of the even-even part of the little Dirac operator
*     to the global single-precision vector field v[] and assigns the result
*     to the field w[]
*
*   void Zooinv(complex *v,complex *w)
*     Applies the inverse of the odd-odd part of the little Dirac operator
*     to the global single-precision vector field v[] and assigns the result
*     to the field w[]
*
*   void Zoe(complex *v,complex *w)
*     Applies the odd-even part of the little Dirac operator to the global
*     single-precision vector field v[] and assigns the result to the
*     field w[]
*
*   void Zeo(complex *v,complex *w)
*     Applies the even-odd part of the little Dirac operator to the global
*     single-precision vector field v[] and *subtracts* the result from
*     the field w[]
*
*   void Znohat(complex *v,complex *w)
*     Applies the little Dirac operator to the global single-precision
*     vector field v[] and assigns the result to the field w[]
*
* Notes:
*
* The restriction of the (improved) Wilson-Dirac operator Dw to the deflation
* subspace is referred to as the little Dirac operator. Its matrix elements
* (and those of the even-odd preconditioned operator) are computed by the
* appropriate programs in the module dfl/Zops.c
*
* The global vector fields are assumed to be those allocated by the programs 
* in the module dfl/vflds.c. Input and output fields must be different (the
* behaviour is otherwise undefined)
* 
* All programs in this module perform global operations and should therefore
* be called simultaneously on all processes
*
*******************************************************************************/

#define ZHAT_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "misc.h"
#include "dfl.h"
#include "global.h"

static int Nss=0;
static complex *rs;


static void alloc_rs(int Ns)
{
   if (Nss>0)
      afree(rs);

   rs=amalloc(Ns*sizeof(*rs),ALIGN);
   
   error(rs==NULL,1,"alloc_rs [Zhat.c]",
         "Unable to allocate auxiliary array");

   Nss=Ns;
}


static void apply_zoe(int Ns,int *in,complex *z,complex *v)
{
   int nmat,*im;
   
   cmat_vec(Ns,z,v+(*in)*Ns,rs);

   nmat=Ns*Ns;   
   z+=nmat;
   im=in+8;
   in+=1;

   for (;in<im;in++)
   {
      cmat_vec_assign(Ns,z,v+(*in)*Ns,rs);
      z+=nmat;
   }
}


static void apply_zeo(int Ns,int *in,complex *z,complex *v)
{
   int nmat,*im;
   
   nmat=Ns*Ns;   
   im=in+8;

   for (;in<im;in++)
   {
      cmat_vec_assign(Ns,z,rs,v+(*in)*Ns);
      z+=nmat;
   }
}


void Zhat(complex *v,complex *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex *zoe,*zeo,*r,*rm;
   zop_t *zop;

   zop=zophat();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rs(Ns);
   
   assign_v2v((nb/2)*Ns,v,w);
   
   if (NPROC>1)
   {
      set_v2zero((nbb/2)*Ns,w+nb*Ns);
      cpv_int_bnd(v);
   }
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   zeo=(*zop).zeo;

   for (;in<im;in+=8)
   {
      apply_zoe(Ns,in,zoe,v);

      r=rs;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*r).re=-(*r).re;
         (*r).im=-(*r).im;
      }

      apply_zeo(Ns,in,zeo,w);      

      zoe+=nz;
      zeo+=nz;
   }

   if (NPROC>1)
      cpv_ext_bnd(w);   
}


static void apply_zdiag(int n,int Ns,complex *z,complex *v,
                             complex *w)
{
   int nmat;
   complex *vm;

   nmat=Ns*Ns;
   vm=v+n*Ns;

   for (;v<vm;v+=Ns)
   {
      cmat_vec(Ns,z,v,w);      
      z+=nmat;
      w+=Ns;
   }
}


void Zeeinv(complex *v,complex *w)
{
   int nb,Ns;
   zop_t *zop;

   zop=zophat();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   
   apply_zdiag(nb/2,Ns,(*zop).zee,v,w);
}


void Zooinv(complex *v,complex *w)
{
   int nb,Ns,n;
   zop_t *zop;

   zop=zophat();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   n=(nb/2)*Ns;
   
   apply_zdiag(nb/2,Ns,(*zop).zoo,v+n,w+n);
}


void Zoe(complex *v,complex *w)
{
   int nb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex *zoe,*r,*rm,*rw;
   zop_t *zop;

   zop=zopnohat();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rs(Ns);
   
   if (NPROC>1)
      cpv_int_bnd(v);
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   rw=w+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      apply_zoe(Ns,in,zoe,v);

      r=rs;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*rw).re=(*r).re;
         (*rw).im=(*r).im;
         rw+=1;
      }

      zoe+=nz;
   }
}


void Zeo(complex *v,complex *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex *zeo,*r,*rm,*rv;
   zop_t *zop;

   zop=zopnohat();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rs(Ns);
   
   if (NPROC>1)
      set_v2zero((nbb/2)*Ns,w+nb*Ns);
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zeo=(*zop).zeo;
   rv=v+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      r=rs;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*r).re=-(*rv).re;
         (*r).im=-(*rv).im;
         rv+=1;
      }
      
      apply_zeo(Ns,in,zeo,w);

      zeo+=nz;
   }

   if (NPROC>1)
      cpv_ext_bnd(w);    
}


void Znohat(complex *v,complex *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex *zoe,*zeo,*r,*rm,*rv,*rw;
   zop_t *zop;

   zop=zopnohat();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rs(Ns);
   
   apply_zdiag(nb,Ns,(*zop).zee,v,w);
   
   if (NPROC>1)
   {
      set_v2zero((nbb/2)*Ns,w+nb*Ns);      
      cpv_int_bnd(v);
   }
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   zeo=(*zop).zeo;
   rv=v+(nb/2)*Ns;
   rw=w+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      apply_zoe(Ns,in,zoe,v);

      r=rs;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*rw).re+=(*r).re;
         (*rw).im+=(*r).im;

         (*r).re=(*rv).re;
         (*r).im=(*rv).im;
         
         rw+=1;
         rv+=1;
      }
      
      apply_zeo(Ns,in,zeo,w);      

      zoe+=nz;
      zeo+=nz;
   }

   if (NPROC>1)
      cpv_ext_bnd(w);   
}

