
/*******************************************************************************
*
* File dfl_modes.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the global modes used to construct the deflation subspace
*
* The externally accessible functions are
*
*   void dfl_modes(int ipr,int *status,int *ifail)
*     Computes the basis vectors of the deflation subspace by applying a
*     smoothing procedure to a set of random fields. The subspace is then
*     initialized using the program dfl_subspace().
*      On exit status reports the average number of solver iterations that
*     were required for the solution of the little Dirac equation in the
*     course of the smoothing procedure (or a negative number if the solver
*     failed) and ifail indicates whether the required inversions of the 
*     SW-term were safe or not (see the notes). A progress report is printed
*     to stdout from process 0 if ipr=1
*
*   void dfl_update(int nmx,int ipr,int *status,int *ifail)
*     Updates the deflation subspace by applying nmx smoothing steps to
*     the global fields psi[i], i=0..,Ns-1, from which the subspace was
*     built. If nmx<=0 the program declares the deflation subspace to be
*     up-to-date (if it is not already so), but leaves the basis vectors
*     unchanged. Otherwise the smoothing procedure is applied and the
*     subspace is constructed from the smoothed fields.
*      On exit status reports the average number of solver iterations that
*     were required for the solution of the little Dirac equation in the
*     course of the smoothing procedure (or a negative number if the solver
*     failed) and ifail indicates whether the required inversions of the 
*     SW-term were safe or not (see the notes). A progress report is printed
*     to stdout from process 0 if ipr=1
*
* Notes:
*
* The deflation subspace is spanned by the fields (*b).sd[1],..,(*b).sd[Ns]
* on the blocks b of the DFL_BLOCKS grid. The number Ns of fields is set by
* the program dfl_set_parms() [flags/parms.c] and it is taken for granted
* that this program is called before dfl_modes(). The program dfl_update()
* moreover assumes that the deflation subspace was previously initialized
* using dfl_modes()
*
* The smoothing procedure operates on a set of Ns global spinor fields. Each
* step consists in the application of a few cycles of the Schwarz alternating
* procedure [sap_gcr/msap.c], optionally deflated using the current set of
* fields. The fields are renormalized after each smoothing step but they are
* not orthogonalized. A Gram-Schmidt orthogonalization is however applied 
* in the program dfl_update() before the smoothing starts. The cycle numbers
* and all other parameters are queried from the parameter data base
*
* In the course of these computations, the SW-term is assigned to the blocks
* of the GCR_BLOCKS grid, first at kappa=dfl.rkappa and finally at kappa=
* lat.kappa, where dfl and lat are the structures returned by dfl_parms() and
* lat_parms() [flags/parms.c]. The assignment inverts the SW term on the odd
* points and, depending on whether any of these fail, the parameter ifail is
* set to
*
*  ifail=0         no failures
*  ifail=1         some inversions at kappa=dfl.kappa failed
*  ifail=2         some inversions at kappa=lat.rkappa failed
*  ifail=3         some inversions failed at both values of kappa
*
* If ifail>=2, the programs skip the smoothing steps. The deflation subspace
* is then still correctly initialized but may have a poor deflation efficiency
*
* The programs dfl_modes() and dfl_update() perform global operations and
* must be called simultaneously on all processes. A workspace of Ns+2 single-
* precision spinor fields is reserved internally using the utilities provided
* by the module flags/wspace.c (these must therefore also be used in the
* calling programs)
*
*******************************************************************************/

#define DFL_MODES_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "random.h"
#include "flags.h"
#include "linalg.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "global.h"

#if (defined SSE)
#include "sse.h"
static sse_float tmp;
#endif

static int my_rank,nvec;
static int Nss=0,rinv,rnmr,rncy,rvnmx;
static int iw0,iw1,*is;
static double rvres,dm;
static complex **vm,**wv,*cs1,*cs2;
static complex_dble **wvd;
static spinor **mds,**ws;


static void print_res(void)
{
   int i;
   double msq;

   for (i=0;i<Nss;i++)
   {
      Qnohat(is[i],iw0);
      msq=(double)(norm_square(VOLUME,1,ws[0])/
                   norm_square(VOLUME,1,mds[i]));

      if (my_rank==0)
         printf("i=%2d, <Q^2> = %.1e\n",i,msq);
   }
}


static void alloc_arrays(int Ns)
{
   if (Nss>0)
   {
      afree(is);
      afree(vm);
      afree(cs1);
   }

   if (Ns>0)
   {
      is=amalloc(Ns*sizeof(*is),3);
      vm=amalloc(Ns*sizeof(*vm),3);
      cs1=amalloc(2*Ns*sizeof(*cs1),3);      
      error((is==NULL)||(vm==NULL)||(cs1==NULL),1,"alloc_arrays [dfl_modes.c]",
            "Unable to allocate auxiliary arrays");
      cs2=cs1+Ns;
      Nss=Ns;
   }
   else
   {
      is=NULL;
      vm=NULL;
      cs1=NULL;
      cs2=NULL;
      Nss=0;
   }
}


static void shift_sw_term(void)
{
   pauli *ms;
   pauli_dble *md,*mm;

   ms=sw;
   md=swd;
   mm=swd+2*VOLUME;

   for (;md<mm;)
   {
      (*md).u[0]+=dm;
      (*md).u[1]+=dm;
      (*md).u[2]+=dm;
      (*md).u[3]+=dm;
      (*md).u[4]+=dm;
      (*md).u[5]+=dm;         

      (*ms).u[0]=(float)((*md).u[0]);
      (*ms).u[1]=(float)((*md).u[1]);
      (*ms).u[2]=(float)((*md).u[2]);
      (*ms).u[3]=(float)((*md).u[3]);
      (*ms).u[4]=(float)((*md).u[4]);      
      (*ms).u[5]=(float)((*md).u[5]);

      md+=1;
      ms+=1;

      (*md).u[0]-=dm;
      (*md).u[1]-=dm;
      (*md).u[2]-=dm;
      (*md).u[3]-=dm;
      (*md).u[4]-=dm;
      (*md).u[5]-=dm;         

      (*ms).u[0]=(float)((*md).u[0]);
      (*ms).u[1]=(float)((*md).u[1]);
      (*ms).u[2]=(float)((*md).u[2]);
      (*ms).u[3]=(float)((*md).u[3]);
      (*ms).u[4]=(float)((*md).u[4]);      
      (*ms).u[5]=(float)((*md).u[5]);

      md+=1;
      ms+=1;
   }
}


static void init_frame(int *ifail)
{
   int nb,isw;
   lat_parms_t lat;
   dfl_parms_t dfl;
   
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   blk_list(GCR_BLOCKS,&nb,&isw);
   
   if (query_flags(U_MATCH_UD)!=1)
      assign_ud2u();

   if (query_blk_flags(GCR_BLOCKS,UBLK_MATCH_UD,nb)!=1)
      assign_ud2ubgr(GCR_BLOCKS);
   
   if ((query_flags(SWD_UP2DATE)!=1)||(query_flags(SWD_E_INVERTED)!=0)||
       (query_flags(SWD_O_INVERTED)!=0))
      sw_term();

   if ((query_flags(SW_UP2DATE)!=1)||(query_flags(SW_E_INVERTED)!=0)||
       (query_flags(SW_O_INVERTED)!=0))
      assign_swd2sw();

   lat=lat_parms();
   dfl=dfl_parms();
   dm=0.5/dfl.rkappa-0.5/lat.kappa;

   if (dm!=0.0)
      shift_sw_term();

   if ((dm!=0.0)||(query_blk_flags(GCR_BLOCKS,SW_UP2DATE,nb)!=1)||
       (query_blk_flags(GCR_BLOCKS,SW_E_INVERTED,nb)!=0)||
       (query_blk_flags(GCR_BLOCKS,SW_O_INVERTED,nb)!=1))
   {
      (*ifail)=2*assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   }
   else
      (*ifail)=0;
   
   blk_list(DFL_BLOCKS,&nb,&isw);
   
   rinv=dfl.rinv;
   rnmr=dfl.rnmr;
   rncy=dfl.rncy;
   rvnmx=dfl.rvnmx;
   rvres=dfl.rvres;
   nvec=nb*dfl.Ns;

   if (dfl.Ns!=Nss)
      alloc_arrays(dfl.Ns);
}


static void random_fields(void)
{
   int i,ix,ic;
   float r,*s;

   r=(float)(1.0/sqrt(24.0*(double)(VOLUME)));
   
   for (i=0;i<Nss;i++)
   {
      s=(float*)(mds[i]);
      
      for (ix=0;ix<VOLUME;ix++)
      {
         ranlxs(s,24);

         for (ic=0;ic<24;ic++)
         {
            (*s)=r*((*s)-0.5f);
            s+=1;
         }
      }
   }
}


static void sum_vprod(int n)
{
   int i;
   
   if (NPROC>1)
   {
      MPI_Reduce((float*)(cs1),(float*)(cs2),2*n,MPI_FLOAT,
                 MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast((float*)(cs2),2*n,MPI_FLOAT,0,MPI_COMM_WORLD);
   }
   else
   {
      for (i=0;i<n;i++)
      {
         cs2[i].re=cs1[i].re;
         cs2[i].im=cs1[i].im;
      }
   }
}


static void restore_fields(void)
{
   int i,j;
   complex z;
   
   dfl_global_modes(vm);

   for (i=0;i<Nss;i++)
   {
      for (j=0;j<i;j++)
         cs1[j]=vprod(nvec,0,vm[j],vm[i]);

      sum_vprod(i);

      for (j=0;j<i;j++)
      {
         z.re=-cs2[j].re;
         z.im=-cs2[j].im;

         mulc_vadd(nvec,vm[i],vm[j],z);
      }

      vnormalize(nvec,1,vm[i]);      
      dfl_v2s(vm[i],is[i]);
   }
   
   set_flags(NEW_DFLMODES);   
}


static void fail_safe(void)
{
   int i;

   for (i=0;i<Nss;i++)
      normalize(VOLUME,1,mds[i]);
   
   dfl_subspace(mds);
}

#if (defined SSE)

static double add_little(int i)
{
   int n;
   float r;
   double x,y;
   spinor *s,*sm,*w;
   
   x=0.0;
   w=ws[1];
   s=mds[i];
   sm=s+VOLUME;

   for (;s<sm;)
   {
      __asm__ __volatile__ ("xorps %%xmm6, %%xmm6 \n\t"
                            "xorps %%xmm7, %%xmm7"
                            :
                            :
                            :
                            "xmm6", "xmm7");
      
      for (n=0;n<4;n++)
      {
         __asm__ __volatile__ ("movaps %0, %%xmm0 \n\t"
                               "movaps %1, %%xmm1 \n\t"
                               "movaps %2, %%xmm2 \n\t"
                               "movaps %3, %%xmm3 \n\t"
                               "movaps %4, %%xmm4 \n\t"
                               "movaps %5, %%xmm5"
                               :
                               :
                               "m" ((*s).c1.c1),
                               "m" ((*s).c1.c3),
                               "m" ((*s).c2.c2),
                               "m" ((*s).c3.c1),
                               "m" ((*s).c3.c3),
                               "m" ((*s).c4.c2)
                               :
                               "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

         __asm__ __volatile__ ("addps %0, %%xmm0 \n\t"
                               "addps %1, %%xmm1 \n\t"
                               "addps %2, %%xmm2 \n\t"
                               "addps %3, %%xmm3 \n\t"
                               "addps %4, %%xmm4 \n\t"
                               "addps %5, %%xmm5"
                               :
                               :
                               "m" ((*w).c1.c1),
                               "m" ((*w).c1.c3),
                               "m" ((*w).c2.c2),
                               "m" ((*w).c3.c1),
                               "m" ((*w).c3.c3),
                               "m" ((*w).c4.c2)
                               :
                               "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

         __asm__ __volatile__ ("movaps %%xmm0, %0 \n\t"
                               "movaps %%xmm1, %1 \n\t"
                               "movaps %%xmm2, %2 \n\t"
                               "movaps %%xmm3, %3 \n\t"
                               "movaps %%xmm4, %4 \n\t"
                               "movaps %%xmm5, %5"
                               :
                               "=m" ((*s).c1.c1),
                               "=m" ((*s).c1.c3),
                               "=m" ((*s).c2.c2),
                               "=m" ((*s).c3.c1),
                               "=m" ((*s).c3.c3),
                               "=m" ((*s).c4.c2));

         s+=4;
         _prefetch_spinor(s);
         s-=3;      
         
         __asm__ __volatile__ ("mulps %%xmm0, %%xmm0 \n\t"
                               "mulps %%xmm1, %%xmm1 \n\t"
                               "mulps %%xmm2, %%xmm2 \n\t"
                               "mulps %%xmm3, %%xmm3 \n\t"
                               "mulps %%xmm4, %%xmm4 \n\t"
                               "mulps %%xmm5, %%xmm5"
                               :
                               :
                               :
                               "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

         w+=4;
         _prefetch_spinor(w);
         w-=3; 
         
         __asm__ __volatile__ ("addps %%xmm0, %%xmm1 \n\t"
                               "addps %%xmm2, %%xmm3 \n\t"
                               "addps %%xmm4, %%xmm6 \n\t"
                               "addps %%xmm5, %%xmm7 \n\t"
                               "addps %%xmm1, %%xmm6 \n\t"
                               "addps %%xmm3, %%xmm7"
                               :
                               :
                               :
                               "xmm1", "xmm3", "xmm6", "xmm7");
      }

      __asm__ __volatile__ ("addps %%xmm6, %%xmm7 \n\t"
                            "movaps %%xmm7, %0"
                            :
                            "=m" (tmp)
                            :
                            :
                            "xmm7");

      x+=(double)(tmp.c1+tmp.c2+tmp.c3+tmp.c4);
   }

   if (NPROC>1)
   {
      MPI_Reduce(&x,&y,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&y,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      x=y;
   }

   if (x!=0.0);
   {
      r=(float)(1.0/sqrt(x));
      
      __asm__ __volatile__ ("movss %0, %%xmm6 \n\t"
                            "shufps $0x0, %%xmm6, %%xmm6 \n\t"
                            "movaps %%xmm6, %%xmm7"
                            :
                            :
                            "m" (r)
                            :
                            "xmm6", "xmm7");      

      s=mds[i];
      sm=s+VOLUME;

      for (;s<sm;s++)
      {
         __asm__ __volatile__ ("movaps %0, %%xmm0 \n\t"
                               "movaps %1, %%xmm1 \n\t"
                               "movaps %2, %%xmm2 \n\t"
                               "movaps %3, %%xmm3 \n\t"
                               "movaps %4, %%xmm4 \n\t"
                               "movaps %5, %%xmm5"
                               :
                               :
                               "m" ((*s).c1.c1),
                               "m" ((*s).c1.c3),
                               "m" ((*s).c2.c2),
                               "m" ((*s).c3.c1),
                               "m" ((*s).c3.c3),
                               "m" ((*s).c4.c2)
                               :
                               "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

         __asm__ __volatile__ ("mulps %%xmm6, %%xmm0 \n\t"
                               "mulps %%xmm7, %%xmm1 \n\t"
                               "mulps %%xmm6, %%xmm2 \n\t"
                               "mulps %%xmm7, %%xmm3 \n\t"
                               "mulps %%xmm6, %%xmm4 \n\t"
                               "mulps %%xmm7, %%xmm5"
                               :
                               :
                               :
                               "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");
      
         __asm__ __volatile__ ("movaps %%xmm0, %0 \n\t"
                               "movaps %%xmm1, %1 \n\t"
                               "movaps %%xmm2, %2 \n\t"
                               "movaps %%xmm3, %3 \n\t"
                               "movaps %%xmm4, %4 \n\t"
                               "movaps %%xmm5, %5"
                               :
                               "=m" ((*s).c1.c1),
                               "=m" ((*s).c1.c3),
                               "=m" ((*s).c2.c2),
                               "=m" ((*s).c3.c1),
                               "=m" ((*s).c3.c3),
                               "=m" ((*s).c4.c2));
      }
   }

   return x;
}

#else

static double add_little(int i)
{
   complex z;

   z.re=1.0f;
   z.im=0.0f;
   mulc_spinor_add(VOLUME,mds[i],ws[1],z);

   return (double)(normalize(VOLUME,1,mds[i]));
}

#endif

static void smooth_field(int i,int ncy,int ipr)
{
   int n,isr;
   double r0,r1;

   set_s2zero(VOLUME,ws[0]);
   isr=is[i];
   
   if (ipr==1)
   {
      r0=(double)(norm_square(VOLUME,1,mds[i]));

      for (n=0;n<ncy;n++)
         msap(rnmr,isr,iw0);      

      r1=(double)(norm_square(VOLUME,1,mds[i]));
      r0=sqrt(r0);
      r1=sqrt(r1);

      if (my_rank==0)
         printf("Smoothing of mode no %2d using Msap: residue %.1e -> %.1e, "
                "ratio = %.1e\n",i,r0,r1,r1/r0);      
   }
   else
   {
      for (n=0;n<ncy;n++)
         msap(rnmr,isr,iw0);
   }

   assign_s2s(VOLUME,ws[0],mds[i]);
}


static void dfl_smooth_field(int i,int ncy,int ipr,int *status)
{
   int n,isr;
   double r0,r1;

   assign_v2vd(nvec,vm[i],wvd[0]);
   dfl_solve_little(rvnmx,rvres,wvd[0],status);
   assign_vd2v(nvec,wvd[0],wv[0]);

   dfl_v2s(wv[0],iw1);
   Qnohat(iw1,iw0);
   mulg5_spinor_sub(VOLUME,mds[i],ws[0]);
   
   set_s2zero(VOLUME,ws[0]);
   isr=is[i];
   
   if (ipr==1)
   {
      r0=(double)(norm_square(VOLUME,1,mds[i]));

      for (n=0;n<ncy;n++)
         msap(rnmr,isr,iw0);      

      r1=(double)(norm_square(VOLUME,1,mds[i]));
      r0=sqrt(r0);
      r1=sqrt(r1);

      if (my_rank==0)
         printf("Deflated smoothing of mode no %2d using Msap: status = %d, "
                "residue %.1e -> %.1e, ratio = %.1e\n",i,*status,r0,r1,r1/r0);
   }
   else
   {
      for (n=0;n<ncy;n++)
         msap(rnmr,isr,iw0);
   }

   assign_s2s(VOLUME,ws[0],mds[i]);
   add_little(i);
}


void dfl_modes(int ipr,int *status,int *ifail)
{
   int iprms[1],n,nh,i,stat;

   if (NPROC>1)
   {
      iprms[0]=ipr;

      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);

      error(iprms[0]!=ipr,1,"dfl_update [dfl_modes.c]",
            "Parameters are not global");
   }

   (*status)=0;   
   init_frame(ifail);
   
   if ((my_rank==0)&&(ipr==1))
   {
      printf("Progress report [program dfl_modes]:\n\n");
      printf("Ns = %d, rinv = %d, rnmr = %d, rncy = %d, rvnmx = %d, "
             "rvres = %.1e\n",Nss,rinv,rnmr,rncy,rvnmx,rvres);
      printf("dm = %.2e, ifail = %d\n\n",dm,*ifail);
   }

   mds=reserve_ws(Nss);
   ws=reserve_ws(2);
   wvd=reserve_wvd(1);
   wv=reserve_wv(1);

   for (i=0;i<Nss;i++)
      is[i]=(mds[i]-ps[0][0])/NSPIN;
   
   iw0=(ws[0]-ps[0][0])/NSPIN;
   iw1=(ws[1]-ps[0][0])/NSPIN;
   random_fields();
   
   if ((*ifail)==0)
   {
      for (n=0;n<3;n++)
      {
         for (i=0;i<Nss;i++)
            smooth_field(i,n+1,ipr);

         if (ipr==1)
            print_res();
      }
      
      for (;n<rinv;n++)
      {
         dfl_subspace(mds);
         dfl_global_modes(vm);

         for (i=0;i<Nss;i++)
         {
            dfl_smooth_field(i,rncy,ipr,&stat);

            if (stat>=0)
               (*status)+=stat;
            else
            {
               (*status)=stat;
               break;
            }
         }

         if ((*status)<0)
            break;

         if (ipr==1)
            print_res();            
      }

      if ((*status)>=0)
      {
         dfl_subspace(mds);

         n=Nss*(rinv-3);
         nh=n/2;
         (*status)=((*status)+nh)/n;
      }
      else
         fail_safe();
   }
   else
      fail_safe();
   
   if (dm!=0.0)
   {
      dm=-dm;
      shift_sw_term();
      (*ifail)+=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   }

   release_wv();
   release_wvd();
   release_ws();
   release_ws();
   
   if ((my_rank==0)&&(ipr==1))
   {
      printf("status = %d, ifail = %d\n",(*status),(*ifail));
      printf("dfl_modes: all done\n\n");
      fflush(stdout);
   }
}


void dfl_update(int nmx,int ipr,int *status,int *ifail)
{
   int iprms[2],n,nh;
   int i,stat;

   if (NPROC>1)
   {
      iprms[0]=nmx;
      iprms[1]=ipr;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=nmx)||(iprms[1]!=ipr),1,"dfl_update [dfl_modes.c]",
            "Parameters are not global");
   }

   error(query_flags(DFLMODES_INITIALIZED)!=1,1,"dfl_update [dfl_modes.c]",
         "The deflation subspace is not initialized");
   
   (*status)=0;
   (*ifail)=0;
   
   if (nmx<=0)
   {
      if (query_flags(DFLMODES_UP2DATE)!=1)
         set_flags(NEW_DFLMODES);
      return;
   }

   init_frame(ifail);
   
   if ((my_rank==0)&&(ipr==1))
   {
      printf("Progress report [program dfl_update]:\n\n");
      printf("nmx = %d\n",nmx);
      printf("Ns = %d, rinv = %d, rnmr = %d, rncy = %d, rvnmx = %d, "
             "rvres = %.1e\n",Nss,rinv,rnmr,rncy,rvnmx,rvres);
      printf("dm = %.2e, ifail = %d\n\n",dm,*ifail);
   }

   mds=reserve_ws(Nss);
   ws=reserve_ws(2);
   wvd=reserve_wvd(1);
   wv=reserve_wv(1);

   for (i=0;i<Nss;i++)
      is[i]=(mds[i]-ps[0][0])/NSPIN;
   
   iw0=(ws[0]-ps[0][0])/NSPIN;
   iw1=(ws[1]-ps[0][0])/NSPIN;
   restore_fields();
   
   if (ipr==1)
      print_res();    
   
   if ((*ifail)==0)
   {
      for (n=0;n<nmx;n++)
      {
         if (n>0)
            dfl_global_modes(vm);
         
         for (i=0;i<Nss;i++)
         {
            dfl_smooth_field(i,rncy,ipr,&stat);

            if (stat>=0)
               (*status)+=stat;
            else
            {
               (*status)=stat;
               break;
            }
         }
            
         if ((*status)>=0)
            dfl_subspace(mds);
         else
            break;

         if (ipr==1)
            print_res();
      }
   }

   if ((*status)>0)
   {
      n=nmx*Nss;
      nh=n/2;
      (*status)=((*status)+nh)/n;
   }
   
   if (dm!=0.0)
   {
      dm=-dm;
      shift_sw_term();
      (*ifail)+=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   }

   release_wv();
   release_wvd();
   release_ws();
   release_ws();
   
   if ((my_rank==0)&&(ipr==1))
   {
      printf("status = %d, ifail = %d\n",(*status),(*ifail));
      printf("dfl_update: all done\n\n");
      fflush(stdout);
   }
}
