
/*******************************************************************************
*
* File parms.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Parameter data base
*
* The externally accessible functions are
*
*   lat_parms_t set_lat_parms(double beta,double kappa,double csw)
*     Sets the basic lattice parameters. The return value is a structure
*     that contains the lattice parameters
*
*   lat_parms_t lat_parms(void)
*     Returns the current lattice parameters in a structure with elements
*     *.beta,*.kappa,*.m0,*.csw where m0 is bare quark mass determined by
*     kappa
*
*   gcr_parms_t set_gcr_parms(int *bs,int nmr,int ncy,int nkv)
*     Sets the parameters for the Schwarz-preconditioned GCR solver and
*     allocates (or reallocates) the GCR_BLOCKS grid. Here bs[4] is the
*     block size, nmr the number of block MR iterations, ncy the number
*     of Schwarz cycles and nkv the maximal number of Krylov vectors to
*     be used before the GCR algorithm is restarted. The return value is
*     a structure that contains the solver parameters
*
*   gcr_parms_t gcr_parms(void)
*     Returns the parameters currently set for the Schwarz-preconditioned
*     GCR solver in a structure with elements *.nb,*.nmr,*.ncy,*.nkv where
*     nb is the number of blocks per process in the current GCR_BLOCKS grid
*
*   hmc_parms_t set_hmc_parms(int *bs,int bc,int N0,int N1,int N2,
*                             int dfl,int cno0,int cno1,int cno_blk,
*                             int nmx_acf,int nmx_gcr_acf,int nmx_rop,
*                             int nmx_frf,int nmx_gcr_frf,
*                             double tau,
*                             double res_acf,double res_gcr_acf,double res_rop,
*                             double res_frf,double res_gcr_frf)
*     Sets the parameters of the Schwarz-preconditioned HMC algorithm and
*     and allocates (or reallocates) the HMC_BLOCKS grid. The parameters are
*
*       bs[4]         Block size of the blocks in the HMC_BLOCKS grid
*
*       bc            Sets the boundary conditions of the quark fields in the
*                     time direction (bc=1: periodic, bc=-1: antiperiodic)
*
*       dfl           Specifies whether the SAP+GCR solver should be deflated
*                     (dfl=1) or not (dfl=0)
* 
*       cno0,cno1     The solutions of the (global) Dirac equation are
*                     propagated along the HMD trajectories, using cno0 and
*                     cno1 previous solutions in the case of the first and
*                     the second Dirac equation solved in hmc_gcr_vec()
*
*       cno_blk       Same as cno0,cno1, but for the case of the solutions of
*                     the normal equation on the blocks of the HMC_BLOCKS grid
*
*       N0,N1,N2      Integration step numbers (as in the CPC paper)
*
*       tau           Trajectory length
*
*       nmx_acf       Maximal solver iteration number and required relative
*       res_acf       residue when calculating the block pseudo-fermion action
*
*       nmx_gcr_acf   Same for the calculation of the block-interaction
*       res_gcr_acf   pseudo-fermion action ("gcr" = global correction)
*
*       nmx_rop       Same for the evaluation of the R-operator, when
*       res_rop       calculating the block-interaction pseudo-fermion field
*                     at the beginning of the trajectory
*
*       nmx_frf       Same for the calculation of block force         
*       res_frf 
*
*       nmx_gcr_frf   Same for the calculation of the block-interaction force
*       res_gcr_frf
*
*     The return value is a structure that contains the HMC parameters
*
*   hmc_parms_t hmc_parms(void)
*     Returns the parameters currently set for the Schwarz-preconditioned
*     HMC algorithm in a structure with elements *.nb,*.bc,*.N0,*.N1,..,
*     where nb is the number of blocks per process in the HMC_BLOCKS grid.
*     The other elements match the arguments of set_hmc_parms() apart from
*     the block size bs which is omitted
*
*   dfl_parms_t set_dfl_parms(int *bs,int Ns,int rinv,int rnmr,int rncy,
*                             int rvnmx,int vnkv,int vnmx,int vdnmx,
*                             double rkappa,double rvres,double vres,
*                             double vdres,int no_v,int no_vd)
*     Sets the parameters of the deflation preconditioner and allocates (or
*     reallocates) the DFL_BLOCKS grid and the associated global complex 
*     fields. The parameters are
*
*       bs[4]         Block size of the blocks in the DFL_BLOCKS grid
*
*       Ns            Dimension of the local subspaces (must be even)
*
*       rkappa        Relaxation hopping parameter
*
*       rinv          Number of inverse iteration steps to be applied when
*                     generating of the deflation modes (rinv>=4)
*
*       rnmr          Number of block MR iterations and number of Schwarz
*       rncy          cycles to be used in the inverse iteration steps
*
*       rvnmx         Maximal solver iteration number and required relative
*       rvres         residue when solving the single-precision little Dirac
*                     equation in the bootstrapped relaxation process
*       
*       vnkv          Maximal number of Krylov vectors to be used by the 
*                     solver for the little Dirac equation before a restart
*
*       vnmx          Maximal solver iteration number and required relative
*       vres          residue when solving the single-precision little Dirac
*                     equation
*
*       vdnmx         Maximal solver iteration number and required relative
*       vdres         residue when solving the double-precision little Dirac
*                     equation
*
*       no_v          Number of global single-precision complex fields to
*                     be allocated in addition to those used internally
*
*       no_vd         Number of global double-precision complex fields to
*                     be allocated in addition to those used internally
*
*     The return value is a structure that contains the DFL parameters
*
*   void set_dfl_residues(double vres,double vdres)
*     Resets the relative residues required when solving the single- and
*     double-precision little Dirac equation to the specified values
*
*   dfl_parms_t dfl_parms(void)
*     Returns the parameters currently set for the deflation preconditioner
*     in a structure with the elements listed above (bs, no_v and no_vd are
*     omitted)
*
* Notes:
*
* To ensure the consistency of the data base, the parameters must be set
* simultaneously on all processes. The types lat_parms_t, ... are defined
* in the file flags.h
*
*******************************************************************************/

#define PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "start.h"
#include "flags.h"
#include "block.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static lat_parms_t lat={0.0,0.0,0.0,0.0};
static gcr_parms_t gcr={0,0,0,0};
static hmc_parms_t hmc={0,1,1,1,1,0,0,0,0,1,1,1,1,1,0.5,1.0,1.0,1.0,1.0,1.0};
static dfl_parms_t dfl={0,1,0,1,1,1,1,0.0,1.0,1.0,1.0,1.0};


lat_parms_t set_lat_parms(double beta,double kappa,double csw)
{
   double dprms[3];

   if (NPROC>1)
   {
      dprms[0]=beta;
      dprms[1]=kappa;
      dprms[2]=csw;

      MPI_Bcast(dprms,3,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((dprms[0]!=beta)||(dprms[1]!=kappa)||(dprms[2]!=csw),1,
             "set_lat_parms [parms.c]","Parameters are not global");
   }

   error_root(kappa<=0.0,1,
              "set_lat_parms [parms.c]","kappa must be positive");

   if (lat.csw!=csw)
      set_flags(CHANGED_CSW);
   else if (lat.kappa!=kappa)
      set_flags(CHANGED_QUARK_MASS);

   lat.beta=beta;
   lat.kappa=kappa;
   lat.m0=1.0/(2.0*kappa)-4.0;
   lat.csw=csw;

   return lat;
}


lat_parms_t lat_parms(void)
{
   return lat;
}


gcr_parms_t set_gcr_parms(int *bs,int nmr,int ncy,int nkv)
{
   int iprms[3],nb,isw;

   if (NPROC>1)
   {
      iprms[0]=nmr;
      iprms[1]=ncy;
      iprms[2]=nkv;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=nmr)||(iprms[1]!=ncy)||(iprms[2]!=nkv),1,
             "set_gcr_parms [parms.c]","Parameters are not global");
   }

   error_root((nmr<0)||(ncy<0)||(nkv<1),1,
               "set_gcr_parms [parms.c]","Improper parameters nmr,ncy,nkv");
   alloc_bgr(GCR_BLOCKS,bs);
   blk_list(GCR_BLOCKS,&nb,&isw);

   gcr.nb=nb;
   gcr.nmr=nmr;
   gcr.ncy=ncy;
   gcr.nkv=nkv;

   return gcr;
}


gcr_parms_t gcr_parms(void)
{
   return gcr;
}


hmc_parms_t set_hmc_parms(int *bs,int bc,int N0,int N1,int N2,
                          int dfl,int cno0,int cno1,int cno_blk,
                          int nmx_acf,int nmx_gcr_acf,int nmx_rop,
                          int nmx_frf,int nmx_gcr_frf,
                          double tau,
                          double res_acf,double res_gcr_acf,double res_rop,
                          double res_frf,double res_gcr_frf)
{
   int iprms[13],nb,isw,icno;
   double dprms[6];

   if (NPROC>1)
   {
      iprms[0]=bc;
      iprms[1]=N0;
      iprms[2]=N1;
      iprms[3]=N2;
      iprms[4]=dfl;
      iprms[5]=cno0;
      iprms[6]=cno1;      
      iprms[7]=cno_blk;      
      iprms[8]=nmx_acf;
      iprms[9]=nmx_gcr_acf;
      iprms[10]=nmx_rop;
      iprms[11]=nmx_frf;      
      iprms[12]=nmx_gcr_frf;

      dprms[0]=tau;
      dprms[1]=res_acf;
      dprms[2]=res_gcr_acf;
      dprms[3]=res_rop;
      dprms[4]=res_frf;
      dprms[5]=res_gcr_frf;
      
      MPI_Bcast(iprms,13,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,6,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=bc)||(iprms[1]!=N0)||(iprms[2]!=N1)||(iprms[3]!=N2)||
            (iprms[4]!=dfl)||(iprms[5]!=cno0)||(iprms[6]!=cno1)||
            (iprms[7]!=cno_blk)||(iprms[8]!=nmx_acf)||(iprms[9]!=nmx_gcr_acf)||
            (iprms[10]!=nmx_rop)||(iprms[11]!=nmx_frf)||
            (iprms[12]!=nmx_gcr_frf)||(dprms[0]!=tau)||(dprms[1]!=res_acf)||
            (dprms[2]!=res_gcr_acf)||(dprms[3]!=res_rop)||
            (dprms[4]!=res_frf)||(dprms[5]!=res_gcr_frf),1,
            "set_hmc_parms [parms.c]","Parameters are not global");
   }

   error_root(((bc!=1)&&(bc!=-1))||(N0<1)||(N1<1)||(N2<1),1,
              "set_hmc_parms [parms.c]","Improper parameters bc or N*");
   error_root((dfl<0)||(dfl>1)||(cno0<0)||(cno1<0)||(cno_blk<0),1,
              "set_hmc_parms [parms.c]","Improper parameters dfl or cno*");
   error_root((nmx_acf<1)||(nmx_gcr_acf<1)||(nmx_rop<1)||
              (nmx_frf<1)||(nmx_gcr_frf<1),1,
              "set_hmc_parms [parms.c]","Improper parameters nmx_*");
   error_root((res_acf<=DBL_EPSILON)||(res_gcr_acf<=DBL_EPSILON)||
              (res_rop<=DBL_EPSILON)||(res_frf<=DBL_EPSILON)||
              (res_gcr_frf<=DBL_EPSILON),1,
              "set_hmc_parms [parms.c]","Improper parameters res_*");

   if ((cno0!=hmc.cno0)||(cno1!=hmc.cno1)||(cno_blk!=hmc.cno_blk))
      icno=1;
   else
      icno=0;
   
   hmc.bc=bc;
   hmc.N0=N0;
   hmc.N1=N1;
   hmc.N2=N2;
   hmc.dfl=dfl;
   hmc.cno0=cno0;
   hmc.cno1=cno1;
   hmc.cno_blk=cno_blk;
   hmc.nmx_acf=nmx_acf;
   hmc.nmx_gcr_acf=nmx_gcr_acf;
   hmc.nmx_rop=nmx_rop;   
   hmc.nmx_frf=nmx_frf;
   hmc.nmx_gcr_frf=nmx_gcr_frf;
   hmc.tau=tau;
   hmc.res_acf=res_acf;
   hmc.res_gcr_acf=res_gcr_acf;   
   hmc.res_rop=res_rop;
   hmc.res_frf=res_frf;
   hmc.res_gcr_frf=res_gcr_frf;

   alloc_bgr(HMC_BLOCKS,bs);
   blk_list(HMC_BLOCKS,&nb,&isw);

   if ((icno==1)||(hmc.nb!=nb))
      alloc_chrono_flds();

   hmc.nb=nb;
   
   return hmc;
}


hmc_parms_t hmc_parms(void)
{
   return hmc;
}


dfl_parms_t set_dfl_parms(int *bs,int Ns,int rinv,int rnmr,int rncy,
                          int rvnmx,int vnkv,int vnmx,int vdnmx,
                          double rkappa,double rvres,double vres,double vdres,
                          int no_v,int no_vd)
{
   int iprms[10];
   double dprms[4];

   if (NPROC>1)
   {
      iprms[0]=Ns;
      iprms[1]=rinv;
      iprms[2]=rnmr;
      iprms[3]=rncy;
      iprms[4]=rvnmx;
      iprms[5]=vnkv;
      iprms[6]=vnmx;
      iprms[7]=vdnmx;
      iprms[8]=no_v;
      iprms[9]=no_vd;

      dprms[0]=rkappa;
      dprms[1]=rvres;
      dprms[2]=vres;
      dprms[3]=vdres;      
      
      MPI_Bcast(iprms,10,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,4,MPI_DOUBLE,0,MPI_COMM_WORLD);
      
      error((iprms[0]!=Ns)||(iprms[1]!=rinv)||(iprms[2]!=rnmr)||
            (iprms[3]!=rncy)||(iprms[4]!=rvnmx)||(iprms[5]!=vnkv)||
            (iprms[6]!=vnmx)||(iprms[7]!=vdnmx)||(iprms[8]!=no_v)||
            (iprms[9]!=no_vd)||(dprms[0]!=rkappa)||(dprms[1]!=rvres)||
            (dprms[2]!=vres)||(dprms[3]!=vdres),1,
            "set_dfl_parms [parms.c]","Parameters are not global");
   }

   free_vbufs();
   free_vdbufs();   
   free_zops();

   error_root((Ns<2)||((Ns%2)!=0),1,"set_dfl_parms [parms.c]",
              "Ns is out of range or not divisible by 2");
   error_root((rinv<4)||(rnmr<1)||(rncy<1)||(rvnmx<1),1,
              "set_dfl_parms [parms.c]",
              "Improper parameter rinv,rnmr,rncy or rvnmx");
   error_root((vnkv<1)||(vnmx<1)||(vdnmx<1),1,"set_dfl_parms [parms.c]",
              "Improper Krylov vector number vnkv, vnmx or vdnmx");
   error_root(rkappa<=0.0,1,"set_dfl_parms [parms.c]",
              "mkappa must be positive");
   error_root((rvres<=(double)(FLT_EPSILON))||(vres<=(double)(FLT_EPSILON))||
              (vdres<=DBL_EPSILON),1,"set_dfl_parms [parms.c]",
              "Improper tolerance mvres, vres or vdres");
   error_root((no_v<0)||(no_vd<0),1,"set_dfl_parms [parms.c]",
              "Improper parameter no_v or no_vd");
   
   dfl.Ns=Ns;
   dfl.rinv=rinv;
   dfl.rnmr=rnmr;
   dfl.rncy=rncy;
   dfl.rvnmx=rvnmx;
   dfl.vnkv=vnkv;
   dfl.vnmx=vnmx;
   dfl.vdnmx=vdnmx;
   dfl.rkappa=rkappa;
   dfl.rvres=rvres;
   dfl.vres=vres;
   dfl.vdres=vdres;   

   alloc_bgr(DFL_BLOCKS,bs);   
   alloc_vflds(2*vnkv+2+no_v);
   alloc_vdflds(4+no_vd);
   
   return dfl;
}


void set_dfl_residues(double vres,double vdres)
{
   dfl.vres=vres;
   dfl.vdres=vdres;    
}


dfl_parms_t dfl_parms(void)
{
   return dfl;
}
