
/*******************************************************************************
*
* File chrono.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs related to the propagation of the solutions of the Dirac equation
* along the molecular-dynamics trajectories
*
* The externally accessible functions are
*
*   void alloc_chrono_flds(void)
*     Allocates the memory space required for the storage of the old
*     solution vectors
*
*   void free_chrono_flds(void)
*     Frees the space previously allocated by alloc_chrono_flds()
*
*   void add_chrono_blk(int n,spinor_dble *sd)
*     Assigns the block spinor field sd to the next field in the stack of
*     old solution vectors on block number n in the HMC_BLOCKS grid
*
*   int old_chrono_blk(int n,double *rd,spinor_dble **sd)
*     Returns an integer nold>=0 and the addresses sd[0],..,sd[nold-1] of
*     a set of orthogonal fields with norm rd[0],..,rd[nold-1] which span
*     the space of the last nold fields on block number n in the HMC_BLOCKS
*     grid. The fields are obtained by reordering the old fields so that
*     the more recent come first and by applying the Gram-Schmidt process
*     (without normalization step) to these
*
*   void reset_chrono_blk(void)
*     Sets the current number of the available old fields on the blocks
*     of the HMC_BLOCKS grid to zero
*
*   void add_chrono(int n,spinor_dble *sd)
*     Assigns the spinor field sd to the next field in the stack of old
*     solution vectors of the first (n=0) or second (n=1) linear system
*     that must be solved in the course of the computation of the global
*     correction force
*
*   int old_chrono(int n,spinor_dble **sd)
*     Returns an integer nold>=0 and the addresses sd[0],..,sd[nold-1] of
*     the last nold solution vectors of the first (n=0) or second (n=1)
*     linear system that must be solved in the course of the computation
*     of the global correction force
*
*   void reset_chrono(void)
*     Sets the current number of the available old fields on the full
*     lattice to zero
*
* Notes:
*
* The solutions of the Dirac equation are propagated following
*
*   R.C. Brower et al., "Chronological inversion method for the Dirac
*   matrix in Hybrid Monte Carlo", Nucl. Phys. B484 (1997) 353
*
* In the case of the block forces the minimial residual method is used,
* while the solutions of the global Dirac equation are propagated using
* a polynomial extrapolation. This module only manages the stack of old
* solutions. The solutions are then propagated in the modules force1.c
* and force2.c
*
* The programs *_blk() do not perform any communications and can be called
* locally, any errors being registered by the error_loc function. All other
* programs must be called simultaneously on all processes

* It is taken for granted that the HMC parameters were set by calling
* set_hmc_parms() [flags/parms.c]. This program automatically sets up the
* field stacks by calling alloc_chrono_flds(). It is then normally not
* required to reallocate them from anywhere else
*
*******************************************************************************/

#define CHRONO_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "block.h"
#include "linalg.h"
#include "global.h"

typedef struct
{
   int vol,nsd,nold,next;
   double *rd;
   spinor_dble **sd;
} fstack_t;

static int nbs;
static fstack_t *fstack=NULL,*fstack_blk=NULL;


static void free_fstack(int n,fstack_t *fst)
{
   int i;

   if (fst[0].nsd!=0)
   {
      afree(fst[0].rd);   
      afree(fst[0].sd[0]);
      afree(fst[0].sd);
   }

   for (i=0;i<n;i++)
   {
      fst[i].vol=0;
      fst[i].nsd=0;
      fst[i].nold=0;
      fst[i].next=0;
      fst[i].rd=NULL;
      fst[i].sd=NULL;
   }
}


void free_chrono_flds(void)
{
   if (fstack!=NULL)
   {
      free_fstack(1,fstack);
      free_fstack(1,fstack+1);
      afree(fstack);
      fstack=NULL;
   }

   if (fstack_blk!=NULL)
   {
      free_fstack(nbs,fstack_blk);
      afree(fstack_blk);
      fstack_blk=NULL;
      nbs=0;
   }  
}


static void alloc_fstack(int n,int vol,int nsd,fstack_t *fst)
{
   int i,j;
   double *rd;
   spinor_dble **sd,*p;

   if (nsd>0)
   {
      rd=amalloc(n*nsd*sizeof(*rd),3);
      sd=amalloc(n*nsd*sizeof(*sd),3);
      p=amalloc(n*nsd*vol*sizeof(*p),ALIGN);

      error((rd==NULL)||(sd==NULL)||(p==NULL),1,"alloc_fstack [chrono.c]",
            "Unable to allocate spinor fields");

      set_sd2zero(n*nsd*vol,p);
   
      for (i=0;i<n;i++)
      {
         fst[i].vol=vol;
         fst[i].nsd=nsd;
         fst[i].nold=0;
         fst[i].next=0;
         fst[i].rd=rd;
         fst[i].sd=sd;
      
         for (j=0;j<nsd;j++)
         {
            rd[j]=0.0;
            sd[j]=p;
            p+=vol;
         }

         rd+=nsd;
         sd+=nsd;
      }
   }
   else
   {
      for (i=0;i<n;i++)
      {
         fst[i].vol=vol;
         fst[i].nsd=0;
         fst[i].nold=0;
         fst[i].next=0;
         fst[i].rd=NULL;
         fst[i].sd=NULL;
      }
   }
}


void alloc_chrono_flds(void)
{
   int nb,isw;
   block_t *b;
   hmc_parms_t hmc;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   
   error(b==NULL,1,"alloc_chrono_flds [chrono.c]",
         "HMC_BLOCKS grid is not allocated");

   free_chrono_flds();
   
   hmc=hmc_parms();

   if ((hmc.cno0>0)||(hmc.cno1>0))
   {
      fstack=amalloc(2*sizeof(*fstack),3);
      error(fstack==NULL,1,"alloc_chrono_flds [chrono.c]",
            "Unable to allocate fstack structures");

      alloc_fstack(1,VOLUME,hmc.cno0,fstack);
      alloc_fstack(1,VOLUME,hmc.cno1,fstack+1);
   }
   
   if (hmc.cno_blk>0)
   {
      fstack_blk=amalloc(nb*sizeof(*fstack_blk),3);
      error(fstack_blk==NULL,1,"alloc_chrono_flds [chrono.c]",
            "Unable to allocate fstack_blk structures");
      
      alloc_fstack(nb,(*b).volh,hmc.cno_blk,fstack_blk);
      nbs=nb;
   }
}


void add_chrono_blk(int n,spinor_dble *sd)
{
   int i,j,is,js;
   double rd,tol;
   complex_dble z;
   fstack_t *fst;

   if (error_loc((fstack_blk==NULL)||(n<0)||(n>=nbs),1,
                 "add_chrono_blk [chrono.c]",
                 "Field stack is not allocated or improper parameter n"))
      return;

   fst=fstack_blk+n;

   if ((*fst).nsd>0)
   {
      is=(*fst).next;   
      assign_sd2sd((*fst).vol,sd,(*fst).sd[is]);
      rd=norm_square_dble((*fst).vol,0,(*fst).sd[is]);
      rd=sqrt(rd);
      (*fst).rd[is]=rd;
      tol=DBL_EPSILON*rd;

      if ((*fst).nold<(*fst).nsd)
         (*fst).nold+=1;
   
      for (i=1;i<(*fst).nold;i++)
      {
         is-=1;
         if (is<0)
            is=(*fst).nsd-1;
      
         js=(*fst).next;

         for (j=0;j<i;j++)
         {
            z=spinor_prod_dble((*fst).vol,0,(*fst).sd[js],(*fst).sd[is]);
            rd=-1.0/((*fst).rd[js]*(*fst).rd[js]);
            z.re*=rd;
            z.im*=rd;
            mulc_spinor_add_dble((*fst).vol,(*fst).sd[is],(*fst).sd[js],z);
         
            js-=1;
            if (js<0)
               js=(*fst).nsd-1;
         }
      
         rd=norm_square_dble((*fst).vol,0,(*fst).sd[is]);
         rd=sqrt(rd);
         (*fst).rd[is]=rd;      

         if (rd<tol)
         {
            (*fst).nold=i;
            break;
         }
      }
   
      (*fst).next+=1;
      
      if ((*fst).next>=(*fst).nsd)
         (*fst).next=0;
   }
}


int old_chrono_blk(int n,double *rd,spinor_dble **sd)
{
   int i,is;
   fstack_t *fst;
   
   if (error_loc((fstack_blk==NULL)||(n<0)||(n>=nbs),1,
                 "old_chrono_blk [chrono.c]",
                 "Field stack is not allocated or improper parameter n"))
      return 0;

   fst=fstack_blk+n;
   is=(*fst).next;
   
   for (i=0;i<(*fst).nold;i++)
   {
      is-=1;
      if (is<0)
         is=(*fst).nsd-1;

      rd[i]=(*fst).rd[is];
      sd[i]=(*fst).sd[is];
   }

   return (*fst).nold;
}


void reset_chrono_blk(void)
{
   int n;
   fstack_t *fst;

   if (fstack_blk!=NULL)
   {
      fst=fstack_blk;

      for (n=0;n<nbs;n++)
      {
         (*fst).nold=0;
         (*fst).next=0;
         fst+=1;
      }
   }
}


void add_chrono(int n,spinor_dble *sd)
{
   fstack_t *fst;

   error((fstack==NULL)||(n<0)||(n>1),1,"add_chrono [chrono.c]",
         "Field stack is not allocated or improper parameter n");

   fst=fstack+n;

   if ((*fst).nsd>0)
   {
      assign_sd2sd((*fst).vol,sd,(*fst).sd[(*fst).next]);

      if ((*fst).nold<(*fst).nsd)
         (*fst).nold+=1;
   
      (*fst).next+=1;
      
      if ((*fst).next>=(*fst).nsd)
         (*fst).next=0;
   }
}


int old_chrono(int n,spinor_dble **sd)
{
   int i,is;
   fstack_t *fst;
   
   error((fstack==NULL)||(n<0)||(n>1),1,"old_chrono [chrono.c]",
         "Field stack is not allocated or improper parameter n");

   fst=fstack+n;

   if ((*fst).nsd>0)
   {
      is=(*fst).next;
   
      for (i=0;i<(*fst).nold;i++)
      {
         is-=1;
         if (is<0)
            is=(*fst).nsd-1;

         sd[i]=(*fst).sd[is];
      }
   }

   return (*fst).nold;
}


void reset_chrono(void)
{
   int n;
   fstack_t *fst;

   if (fstack!=NULL)
   {
      fst=fstack;

      for (n=0;n<2;n++)
      {
         (*fst).nold=0;
         (*fst).next=0;
         fst+=1;
      }
   }
}
