
/*******************************************************************************
*
* File force2.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the force F_2 that derives from the block-interaction part
* of the molecular-dynamics action
*
* The externally accessible functions are
*
*   void hmc_gcr_vec(int k1,int k2,int nmx_dfl,int *stat_dfl,int *status,
*                    int *ifail)
*     Computes the global spinor fields
*
*       psi1=Q^(-1)*Q_dOmega^star*chi and  psi2=Q^(-1)*R_2^(-1)*chi
*
*     and assigns them to psd[k1] and psd[k2] respectively. Depending on
*     the parameters returned by hmc_parms(), the ordinary or the deflated 
*     SAP+GCR solver is used for the solution of the Dirac equation.
*       If deflation is on and if nmx_dfl>=1, the deflation subspace is
*     updated by applying nmx_dfl smoothing steps (see dfl/dfl_modes.c). 
*     The SW-term and the fields on the GCR_BLOCKS grid are updated when
*     needed.
*      The parameter status is assumed to be an array of 3 elements, while
*     stat_dfl points to a single value. If deflation is on, status reports
*     the *sum* of the status values returned by dfl_sap_gcr() (which is
*     called two times in this program) and stat_dfl is set to the status
*     value returned by dfl_update(). The status variables are all set
*     to zero if deflation is off except for status[0], which reports the
*     sum of the status values returned by sap_gcr(). Negative values are
*     returned if the corresponding program failed and a positive value of 
*     ifail indicates that the SW-term could not be safely inverted (ifail
*     is set to zero if all inversions were safe)
*
*   void hmc_gcr_frf(int n,int k1,int k2)
*     Computes the force on the n'th block in the HMC_BLOCKS grid, assuming
*     that the fields psi1 and psi2 defined above are stored in the global
*     arrays psd[k1] and psd[k2]
*
*   void hmc_gcr_acf(int k1,int nmx_dfl,double *act,int *stat_dfl,int *status,
*                    int *ifail)
*     Computes the action of the pseudo-fermion field chi. Depending on 
*     the parameters returned by hmc_parms(), the ordinary or the deflated 
*     SAP+GCR solver is used for the solution of the Dirac equation. The
*     solution is accelerated if psd[k1] contains an accurate approximation
*     to the field psi1 defined above (the program checks whether this is
*     the case). The action is returned as an array act[0],...,act[nb-1],
*     where act[n] is the contribution of n'th block in the HMC_BLOCKS grid
*     and nb the number of blocks in the local lattice.
*       If deflation is on and if nmx_dfl>=1, the deflation subspace is
*     updated by applying nmx_dfl smoothing steps (see dfl/dfl_modes.c). 
*     The SW-term and the fields on the GCR_BLOCKS grid are updated when
*     needed.
*      The parameter status is assumed to be an array of 3 elements, while
*     stat_dfl points to a single value. If deflation is on, status reports
*     the status values returned by dfl_sap_gcr() and stat_dfl is set to
*     the status value returned by dfl_update(). The status variables are
*     all set to zero if deflation is off except for status[0], which reports
*     the sum of the status values returned by sap_gcr(). Negative values are
*     returned if the corresponding program failed and a positive value of 
*     ifail indicates that the SW-term could not be safely inverted (ifail
*     is set to zero if all inversions were safe)
*
* Notes:
*
* All programs in this file except for hmc_gcr_frf() must be called on all
* processes simultaneously. Whether the deflation subspace etc. is in the
* proper condition is checked by querying the flags data base
*
* In all cases the gauge field that is used is the double-precision field
* on the blocks of the HMC_BLOCKS grid. The inactive link variables on the
* blocks must match the ones on the global lattice when psi1 and psi2 or
* the action are calculated
*
* The lattice parameters, maximal iteration numbers, ... are retrieved
* from the parameter data base using the utilities provided by the module
* flags/parms.c. Only the forces on the active links are calculated, and the 
* force field on the inactive links is left untouched. The "chronological
* inversion method" is silently applied in the program hmc_gcr_vec() if the
* parameters hmc.cno0 and hmc.cno1 returned by hmc_parms() are not both
* equal to zero
*
* The programs hmc_gcr_vec() and hmc_gcr_acf() require a workspace of 3 and 
* 2 double-precision spinor fields respectively in addition to the workspace
* needed by the (optionally deflated) SAP+GCR solver. If the chronological
* inverter is switched off, hmc_gcr_vec() only needs 2 fields. The workspace
* is reserved using the utilties in the module flags/wspace.c
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define FORCE2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "hmc.h"
#include "global.h"

#define DEBUG 0

static int nmat=0,nold;
static complex_dble *cvec;
static spinor_dble **sdold;


static void alloc_mat(int n)
{
   if (n<=nmat)
      return;

   if (nmat>0)
   {
      afree(cvec);
      afree(sdold);
   }

   cvec=amalloc(n*sizeof(*cvec),4);
   sdold=amalloc(n*sizeof(*sdold),3);
   
   error((cvec==NULL)||(sdold==NULL),1,
         "alloc_mat [force2.c]","Unable to allocate auxiliary arrays");

   nmat=n;
}


static void lagrange(int n)
{
   int i;

   cvec[0].re=(double)(n);
   cvec[0].im=0.0;
   
   for (i=1;i<n;i++)
   {
      cvec[i].re=-cvec[i-1].re*(double)(n-i)/(double)(i+1);
      cvec[i].im=0.0;
   }
}


static void extrapolate(int n,int ir)
{
   int i;
   spinor_dble *sdr;
   
   lagrange(n);

   sdr=psd[ir][0];
   set_sd2zero(VOLUME,sdr);

   for (i=0;i<n;i++)
      mulc_spinor_add_dble(VOLUME,sdr,sdold[i],cvec[i]);
}


void hmc_gcr_vec(int k1,int k2,int nmx_dfl,int *stat_dfl,int *status,
                 int *ifail)
{
   int iprms[3],n,nb,isw,cno;
   int iwd0,ivd0,ivd1,is,ir;
   int stat1[3],stat2[3],flag0,flag1,flag2;
   double res0,res1;
   complex_dble z;
   spinor_dble **wsd,**vsd;
   block_t *b,*bm;
   hmc_parms_t hmc;

   if (NPROC>1)
   {
      iprms[0]=k1;
      iprms[1]=k2;
      iprms[2]=nmx_dfl;
   
      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k1)||(iprms[1]!=k2)||(iprms[2]!=nmx_dfl),1,
            "hmc_gcr_vec [force2.c]","Parameters are not global");  
   }

   hmc=hmc_parms();
   n=query_flags(BCD_FLIPPED);
   error(((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)),1,
         "hmc_gcr_vec [force2.c]","Boundary conditions do not match data base");
   
   wsd=reserve_wsd(1);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;

   error_root((k1<0)||(k1>=no_sd)||(k2<0)||(k2>=no_sd)||
              (k1==k2)||(k1==iwd0)||(k2==iwd0),1,"hmc_gcr_vec [force2.c]",
              "Improper parameters k1,k2");  

   assign_chi2sd(iwd0);
   set_sd2zero(VOLUME,psd[k2][0]);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,iwd0,0);
      Qbnds_dble(b,0,0);
      sub_assign_sdbb2sd(b,0,k2);
   } 

   mulmg5_dble(VOLUME,psd[k2][0]);
   swap_active(HMC_BLOCKS);

   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
      sw_term();

   if (hmc.cno0>hmc.cno1)
      cno=hmc.cno0;
   else
      cno=hmc.cno1;
   
   if (cno>0)
   {
      alloc_mat(cno);

      if (hmc.cno0>0)
         nold=old_chrono(0,sdold);
      else
         nold=0;
   }
   else
      nold=0;

   if (nold>0)
   {
      vsd=reserve_wsd(2);
      ivd0=(vsd[0]-psd[0][0])/NSPIN;
      ivd1=(vsd[1]-psd[0][0])/NSPIN;

      error_root((k1==ivd0)||(k1==ivd1)||(k2==ivd0)||(k2==ivd1),1,
                 "hmc_gcr_vec [force2.c]","Improper parameters k1,k2");
         
      extrapolate(nold,k1);
      Qnohat_dble(k1,ivd1);
      mulg5_dble(VOLUME,vsd[1]);
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble(VOLUME,vsd[1],psd[k2][0],z);      

      res0=norm_square_dble(VOLUME,1,psd[k2][0]);      
      res1=norm_square_dble(VOLUME,1,vsd[1]);
      res1=sqrt(res1/res0);

      if (res1<1.0)
      {
         is=ivd1;
         ir=ivd0;
      }
      else
      {
         res1=1.0;
         is=k2;
         ir=k1;
      }
   }
   else
   {
      vsd=NULL;
      res1=1.0;
      is=k2;
      ir=k1;
   }

#if (DEBUG>=1)
   message("\nhmc_gcr_vec(), 1st solve: nold = %d\n",nold);
   message("res1 = %.1e\n",res1);
#endif
   
   if (hmc.dfl==1)
      dfl_update(nmx_dfl,DEBUG>=2,stat_dfl,&flag0);
   else
   {
      (*stat_dfl)=0;
      flag0=0;
   }
   
   if (res1>hmc.res_gcr_frf)
   {
      if (hmc.dfl==1)
      {
         if (((*stat_dfl)>=0)&&(flag0==0))
            dfl_sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf/res1,is,ir,
                        stat1,&flag1);
         else
         {
            stat1[0]=0;
            stat1[1]=0;
            stat1[2]=0;
            flag1=0;
         }

         flag1|=flag0;
      }
      else
      {
         sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf/res1,is,ir,
                 stat1,&flag1);
         
         stat1[1]=0;
         stat1[2]=0;
      }
   }
   else
   {
      set_sd2zero(VOLUME,psd[ir][0]);
      
      stat1[0]=0;
      stat1[1]=0;
      stat1[2]=0;
      flag1=0;
   }

   if (nold>0)
   {
      if (ir!=k1)
      {
         z.re=-1.0;
         z.im=0.0;
         mulc_spinor_add_dble(VOLUME,psd[k1][0],vsd[0],z);
      }

      release_wsd();
   }

#if (DEBUG>=1)
   message("Status = %d (little = %d,%d), ifail = %d\n",
           stat1[0],stat1[1],stat1[2],flag1);

   vsd=reserve_wsd(1);
   ivd0=(vsd[0]-psd[0][0])/NSPIN;

   Qnohat_dble(k1,ivd0);
   mulg5_dble(VOLUME,vsd[0]);
   z.re=-1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,vsd[0],psd[k2][0],z);
   res1=norm_square_dble(VOLUME,1,vsd[0])/
      norm_square_dble(VOLUME,1,psd[k2][0]);
   res1=sqrt(res1);

   message("Residue = %.1e (should be <= %.1e)\n",res1,hmc.res_gcr_frf);
   
   if (nold>0)
   {
      message("Approximation quality using n = 1,..,%d old fields: ",nold);

      if (hmc.dfl)
      {
         dfl_Qproj_dble(k1,ivd0);
         res0=norm_square_dble(VOLUME,1,vsd[0]);
      }
      else
         res0=norm_square_dble(VOLUME,1,psd[k1][0]);
      
      for (n=1;n<=nold;n++)
      {
         extrapolate(n,ivd0);
         mulc_spinor_add_dble(VOLUME,vsd[0],psd[k1][0],z);
         if (hmc.dfl)
            dfl_Qproj_dble(ivd0,ivd0);
         res1=norm_square_dble(VOLUME,1,vsd[0]);

         if (n<nold)
            message("%.1e, ",sqrt(res1/res0));
         else
            message("%.1e\n",sqrt(res1/res0));
      }
   }

   release_wsd();
#endif   

   if (((*stat_dfl)>=0)&&(stat1[0]>=0)&&(stat1[2]>=0)&&(flag1==0))
   {
      if ((hmc.cno0>0)&&(stat1[0]>0))
         add_chrono(0,psd[k1][0]);
      
      sub_assign_PdOs(iwd0,k1);
      mulg5_dble(VOLUME,wsd[0]);

      if (hmc.cno1>0)
         nold=old_chrono(1,sdold);
      else
         nold=0;

      if (nold>0)
      {
         vsd=reserve_wsd(2);
         ivd0=(vsd[0]-psd[0][0])/NSPIN;
         ivd1=(vsd[1]-psd[0][0])/NSPIN;

         extrapolate(nold,k2);         
         Qnohat_dble(k2,ivd1);
         mulg5_dble(VOLUME,vsd[1]);
         z.re=-1.0;
         z.im=0.0;
         mulc_spinor_add_dble(VOLUME,vsd[1],wsd[0],z);      

         res0=norm_square_dble(VOLUME,1,wsd[0]);      
         res1=norm_square_dble(VOLUME,1,vsd[1]);
         res1=sqrt(res1/res0);

         if (res1<1.0)
         {
            is=ivd1;
            ir=ivd0;
         }
         else
         {
            res1=1.0;
            is=iwd0;
            ir=k2;
         }
      }
      else
      {
         vsd=NULL;
         res1=1.0;
         is=iwd0;
         ir=k2;
      }

#if (DEBUG>=1)
      message("hmc_gcr_vec(), 2nd solve: nold = %d\n",nold);
      message("res1 = %.1e\n",res1);
#endif
      
      if (res1>hmc.res_gcr_frf)
      {
         if (hmc.dfl==1)
            dfl_sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf/res1,is,ir,
                        stat2,&flag2);
         else
         {
            sap_gcr(hmc.nmx_gcr_frf,hmc.res_gcr_frf/res1,is,ir,
                    stat2,&flag2);

            stat2[1]=0;
            stat2[2]=0;
         }
      }
      else
      {
         set_sd2zero(VOLUME,psd[ir][0]);
      
         stat2[0]=0;
         stat2[1]=0;
         stat2[2]=0;
         flag2=0;
      }   

      if (nold>0)
      {
         if (ir!=k2)
         {
            z.re=-1.0;
            z.im=0.0;
            mulc_spinor_add_dble(VOLUME,psd[k2][0],vsd[0],z);
         }

         release_wsd();
      }

#if (DEBUG>=1)
      message("Status = %d (little = %d,%d), ifail = %d\n",
              stat2[0],stat2[1],stat2[2],flag2);

      vsd=reserve_wsd(1);
      ivd0=(vsd[0]-psd[0][0])/NSPIN;

      Qnohat_dble(k2,ivd0);
      mulg5_dble(VOLUME,vsd[0]);
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble(VOLUME,vsd[0],wsd[0],z);
      res1=norm_square_dble(VOLUME,1,vsd[0])/
         norm_square_dble(VOLUME,1,wsd[0]);
      res1=sqrt(res1);

      message("Residue = %.1e (should be <= %.1e)\n",res1,hmc.res_gcr_frf);

      if (nold>0)
      {
         message("Approximation quality using n = 1,..,%d old fields: ",nold);

         if (hmc.dfl)
         {
            dfl_Qproj_dble(k2,ivd0);
            res0=norm_square_dble(VOLUME,1,vsd[0]);
         }
         else
            res0=norm_square_dble(VOLUME,1,psd[k2][0]);
   
         for (n=1;n<=nold;n++)
         {
            extrapolate(n,ivd0);
            mulc_spinor_add_dble(VOLUME,vsd[0],psd[k2][0],z);
            if (hmc.dfl)
               dfl_Qproj_dble(ivd0,ivd0);
            res1=norm_square_dble(VOLUME,1,vsd[0]);

            if (n<nold)
               message("%.1e, ",sqrt(res1/res0));
            else
               message("%.1e\n",sqrt(res1/res0));
         }
      }
      
      release_wsd();
#endif

      if ((hmc.cno1>0)&&(stat2[0]>0)&&(stat2[2]>=0)&&(flag2==0))
         add_chrono(1,psd[k2][0]);
   }
   else
   {
      stat2[0]=0;
      stat2[1]=0;
      stat2[2]=0;
      flag2=0;
   }
   
   swap_active(HMC_BLOCKS);

   for (n=0;n<3;n++)
   {
      if (stat1[n]<0)
         status[n]=stat1[n];
      else if (stat2[n]<0)
         status[n]=stat2[n];
      else
         status[n]=stat1[n]+stat2[n];
   }
   
   (*ifail)=(flag1|flag2);
   release_wsd();
}


void hmc_gcr_frf(int n,int k1,int k2)
{
   int nb,isw;
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw);

   if (error_loc((b==NULL)||(n<0)||(n>=nb),1,"hmc_gcr_frf [force2.c]",
                 "Block grid is not allocated or block number out of range"))
      return;

   if (error_loc((k1<0)||(k1>=no_sd)||(k2<0)||(k2>=no_sd),1,
                 "hmc_gcr_frf [force2.c]","Improper arguments k1,k2"))
      return;

   b+=n;
   assign_sd2sdb(b,ALL_PTS,k1,0);
   assign_sd2sdb(b,ALL_PTS,k2,1);
   
   if (sw_frc(n,0,1,0)==0)
      hop_frc(n,0,1);
}


static void block_action(int k,double *act)
{
   int nb,isw,nbh;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   nbh=nb/2;

   if (isw==1)
   {
      b+=nbh;
      act+=nbh;
   }

   bm=b+nbh;
   
   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      (*act)=norm_square_dble((*b).vol,0,(*b).sd[0]);
      act+=1;
   }
}


void hmc_gcr_acf(int k1,int nmx_dfl,double *act,int *stat_dfl,int *status,
                 int *ifail)
{
   int iprms[2],n,nb,isw;
   int iwd0,iwd1,flag0;
   double res0,res1,res;
   complex_dble z;
   spinor_dble **wsd;
   block_t *b,*bm;
   hmc_parms_t hmc;

   if (NPROC>1)
   {
      iprms[0]=k1;
      iprms[1]=nmx_dfl;
      
      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k1)||(iprms[1]!=nmx_dfl),1,
            "hmc_gcr_acf [force2.c]","Parameters are not global");
   }
   
   hmc=hmc_parms();
   
   n=query_flags(BCD_FLIPPED);
   error(((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)),1,
         "hmc_gcr_acf [force2.c]","Boundary conditions do not match data base");

   wsd=reserve_wsd(2);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;
   
   error_root((k1<0)||(k1>=no_sd)||(k1==iwd0)||(k1==iwd1),1,
              "hmc_gcr_acf [force2.c]","Improper parameter k1");
   
   assign_chi2sd(iwd0);
   set_sd2zero(VOLUME,wsd[1]);

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,iwd0,0);
      Qbnds_dble(b,0,0);
      sub_assign_sdbb2sd(b,0,iwd1);
   } 
   
   res0=norm_square_dble(VOLUME,1,wsd[1]);
   swap_active(HMC_BLOCKS);

   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
      sw_term();
   
   Qnohat_dble(k1,iwd0);
   z.re=1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,wsd[0],wsd[1],z);
   res1=norm_square_dble(VOLUME,1,wsd[0]);
   res=hmc.res_gcr_acf;

   for (n=0;n<nb;n++)
      act[n]=0.0;
   
   if (res1<res0)
   {
      if (res1>(res0*res*res))
      {
         res*=sqrt(res0/res1);
         mulmg5_dble(VOLUME,wsd[0]);

         if (hmc.dfl==1)
         {
            dfl_update(nmx_dfl,DEBUG>=2,stat_dfl,&flag0);

            if (((*stat_dfl)>=0)&&(flag0==0))
               dfl_sap_gcr(hmc.nmx_gcr_acf,res,iwd0,iwd1,status,ifail);
            else
            {
               status[0]=0;
               status[1]=0;
               status[2]=0;
               (*ifail)=0;
            }

            (*ifail)|=flag0;            
         }
         else
         {
            sap_gcr(hmc.nmx_gcr_acf,res,iwd0,iwd1,status,ifail);

            (*stat_dfl)=0;
            status[1]=0;
            status[2]=0;
         }

         if (((*stat_dfl)>=0)&&(status[0]>=0)&&(status[2]>=0)&&((*ifail)==0))
         {
            mulc_spinor_add_dble(VOLUME,wsd[1],psd[k1][0],z);
            assign_chi2sd(iwd0);   
            sub_assign_PdOs(iwd0,iwd1);
            block_action(iwd0,act);                     
         }
      }
      else
      {
         assign_chi2sd(iwd0);   
         sub_assign_PdOs(iwd0,k1);
         block_action(iwd0,act);         

         (*stat_dfl)=0;
         status[0]=0;
         status[1]=0;
         status[2]=0;
         (*ifail)=0;
      }
   }
   else
   {
      mulmg5_dble(VOLUME,wsd[1]);

      if (hmc.dfl==1)
      {
         dfl_update(nmx_dfl,DEBUG>=2,stat_dfl,&flag0);

         if (((*stat_dfl)>=0)&&(flag0==0))
            dfl_sap_gcr(hmc.nmx_gcr_acf,res,iwd1,iwd0,status,ifail);
         else
         {
            status[0]=0;
            status[1]=0;
            status[2]=0;
            (*ifail)=0;
         }

         (*ifail)|=flag0;            
      }
      else
      {
         sap_gcr(hmc.nmx_gcr_acf,res,iwd1,iwd0,status,ifail);

         (*stat_dfl)=0;
         status[1]=0;
         status[2]=0;
      }

      if (((*stat_dfl)>=0)&&(status[0]>=0)&&(status[2]>=0)&&((*ifail)==0))
      {
         assign_chi2sd(iwd1);   
         sub_assign_PdOs(iwd1,iwd0);
         block_action(iwd1,act);
      }
   }

   swap_active(HMC_BLOCKS);
   release_wsd();
}

