
/*******************************************************************************
*
* File frcfcts.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic functions used for the force calculation
*
* The externally accessible functions are
*
*   void det2u3alg(pauli_dble *m,u3_alg_dble *X)
*     Computes the antihermitian matrices X that are associated to the field
*     variation of the determinant of the sw-term at a given lattice point x,
*     assuming that the inverse of the sw-term at x is stored in the two Pauli
*     matrices m[0] and m[1] (see the notes below for further explanations) 
*
*   void vec2u3alg(spinor_dble *r,spinor_dble *s,u3_alg_dble *X)
*     Computes the antihermitian matrices X that are associated to the field
*     variation of Re(r,M*s), where M denotes the sw-term at a given point x
*     and r,s are spinors at x (see the notes below for further explanations) 
*
*   void (*spin2mat[])(spinor_dble *rx,spinor_dble *ry,
*                      spinor_dble *sx,spinor_dble *sy,su3_dble *u)
*     For given mu=0,1,2,3, the program spin2frc[mu] computes the 3x3 matrix
*     u=tr{gamma_5*(1-gamma_mu)*[(ry x sx^dag)+(sy x rx^dag)]}, where ..x..
*     stands for the tensor product in spinor space and the trace is taken 
*     over the Dirac indices only
*
* Notes:
*
* The elements X of the Lie algebra of U(3) are antihermitian 3x3 matrices
* that are represented by structures X with real entries X.c1,...,X.c9
* through
*
*  X_11=i*X.c1, X_22=i*X.c2, X_33=i*X.c3,
*
*  X_12=X.c4+i*X.c5, X_13=X.c6+i*X.c7, X_23=X.c8+i*X.c9
*
* For all planes (mu,nu)=(0,1),(0,2),(0,3),(1,2),(1,3),(2,3) the programs
* det2u3alg and vec2u3alg compute the matrices
*
*  X=i*tr{sigma_{mu,nu}*gamma_5*(m[0],m[1])}              (program det2u3alg)
*
*  X=i*tr{(sigma_{mu,nu}*gamma5*s)x(r^dagger)+h.c.}       (program vec2u3alg)
*
* where the trace is over the Dirac indices only and "(..)x(..)" stands for
* the tensor product in spinor space. In both cases the matrices for the
* different planes are stored in the output array in the specified order
*
*******************************************************************************/

#define FRCFCTS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "start.h"

#if (defined SSE)
static su3_vector_dble psi1,psi2,chi1,chi2 __attribute__ ((aligned (16)));
static pauli_dble ms[2] __attribute__ ((aligned (16)));
#else
static su3_vector_dble psi1,psi2,chi1,chi2;
static pauli_dble ms[2];
#endif

#define _re(z,w) ((z).re*(w).re+(z).im*(w).im)
#define _im(z,w) ((z).im*(w).re-(z).re*(w).im)


void det2u3alg(pauli_dble *m,u3_alg_dble *X)
{
   double x;
   u3_alg_dble *X0,*X1;
   pauli_dble *m0,*m1;

   m0=m;
   m1=m0+1;
   
   X0=X;
   X1=X+5;

   (*X0).c1=(*m0).u[10]+(*m0).u[10];
   (*X1).c1=-(*X0).c1;
   x=(*m1).u[10]+(*m1).u[10];
   (*X0).c1+=x;
   (*X1).c1+=x;

   (*X0).c2=(*m0).u[20]+(*m0).u[20];
   (*X1).c2=-(*X0).c2;
   x=(*m1).u[20]+(*m1).u[20];
   (*X0).c2+=x;
   (*X1).c2+=x;

   (*X0).c3=(*m0).u[28]+(*m0).u[28];
   (*X1).c3=-(*X0).c3;
   x=(*m1).u[28]+(*m1).u[28];
   (*X0).c3+=x;
   (*X1).c3+=x;

   (*X0).c4=(*m0).u[19]-(*m0).u[13];
   (*X1).c4=-(*X0).c4;
   x=(*m1).u[19]-(*m1).u[13];
   (*X0).c4+=x;
   (*X1).c4+=x;

   (*X0).c5=(*m0).u[12]+(*m0).u[18];
   (*X1).c5=-(*X0).c5;
   x=(*m1).u[12]+(*m1).u[18];
   (*X0).c5+=x;
   (*X1).c5+=x;

   (*X0).c6=(*m0).u[25]-(*m0).u[15];   
   (*X1).c6=-(*X0).c6;
   x=(*m1).u[25]-(*m1).u[15];
   (*X0).c6+=x;
   (*X1).c6+=x;

   (*X0).c7=(*m0).u[14]+(*m0).u[24];
   (*X1).c7=-(*X0).c7;
   x=(*m1).u[14]+(*m1).u[24];
   (*X0).c7+=x;
   (*X1).c7+=x;

   (*X0).c8=(*m0).u[27]-(*m0).u[23];   
   (*X1).c8=-(*X0).c8;
   x=(*m1).u[27]-(*m1).u[23];
   (*X0).c8+=x;
   (*X1).c8+=x;

   (*X0).c9=(*m0).u[22]+(*m0).u[26];
   (*X1).c9=-(*X0).c9;
   x=(*m1).u[22]+(*m1).u[26];
   (*X0).c9+=x;
   (*X1).c9+=x;

   X0=X+1;
   X1=X+4;

   (*X0).c1=-(*m0).u[11]-(*m0).u[11];
   (*X1).c1=(*X0).c1;
   x=(*m1).u[11]+(*m1).u[11];
   (*X0).c1-=x;
   (*X1).c1+=x;

   (*X0).c2=-(*m0).u[21]-(*m0).u[21];
   (*X1).c2=(*X0).c2;
   x=(*m1).u[21]+(*m1).u[21];
   (*X0).c2-=x;
   (*X1).c2+=x;

   (*X0).c3=-(*m0).u[29]-(*m0).u[29];
   (*X1).c3=(*X0).c3;
   x=(*m1).u[29]+(*m1).u[29];
   (*X0).c3-=x;
   (*X1).c3+=x;
   
   (*X0).c4=(*m0).u[18]-(*m0).u[12];
   (*X1).c4=(*X0).c4;
   x=(*m1).u[18]-(*m1).u[12];
   (*X0).c4+=x;
   (*X1).c4-=x;

   (*X0).c5=-(*m0).u[13]-(*m0).u[19];
   (*X1).c5=(*X0).c5;
   x=(*m1).u[13]+(*m1).u[19];
   (*X0).c5-=x;
   (*X1).c5+=x;

   (*X0).c6=(*m0).u[24]-(*m0).u[14];
   (*X1).c6=(*X0).c6;
   x=(*m1).u[24]-(*m1).u[14];
   (*X0).c6+=x;
   (*X1).c6-=x;
   
   (*X0).c7=-(*m0).u[25]-(*m0).u[15];
   (*X1).c7=(*X0).c7;
   x=(*m1).u[25]+(*m1).u[15];
   (*X0).c7-=x;
   (*X1).c7+=x;

   (*X0).c8=(*m0).u[26]-(*m0).u[22];
   (*X1).c8=(*X0).c8;
   x=(*m1).u[26]-(*m1).u[22];
   (*X0).c8+=x;
   (*X1).c8-=x;
   
   (*X0).c9=-(*m0).u[27]-(*m0).u[23];
   (*X1).c9=(*X0).c9;
   x=(*m1).u[27]+(*m1).u[23];
   (*X0).c9-=x;
   (*X1).c9+=x;

   X0=X+2;
   X1=X+3;

   (*X0).c1=(*m0).u[0]-(*m0).u[3];
   (*X1).c1=-(*X0).c1;
   x=(*m1).u[0]-(*m1).u[3];
   (*X0).c1+=x;
   (*X1).c1+=x;

   (*X0).c2=(*m0).u[1]-(*m0).u[4];
   (*X1).c2=-(*X0).c2;
   x=(*m1).u[1]-(*m1).u[4];
   (*X0).c2+=x;
   (*X1).c2+=x;

   (*X0).c3=(*m0).u[2]-(*m0).u[5];
   (*X1).c3=-(*X0).c3;
   x=(*m1).u[2]-(*m1).u[5];
   (*X0).c3+=x;
   (*X1).c3+=x;
   
   (*X0).c4=(*m0).u[31]-(*m0).u[7];
   (*X1).c4=-(*X0).c4;
   x=(*m1).u[31]-(*m1).u[7];
   (*X0).c4+=x;
   (*X1).c4+=x;

   (*X0).c5=(*m0).u[6]-(*m0).u[30];
   (*X1).c5=-(*X0).c5;
   x=(*m1).u[6]-(*m1).u[30];
   (*X0).c5+=x;
   (*X1).c5+=x;

   (*X0).c6=(*m0).u[33]-(*m0).u[9];
   (*X1).c6=-(*X0).c6;
   x=(*m1).u[33]-(*m1).u[9];
   (*X0).c6+=x;
   (*X1).c6+=x;

   (*X0).c7=(*m0).u[8]-(*m0).u[32];
   (*X1).c7=-(*X0).c7;
   x=(*m1).u[8]-(*m1).u[32];
   (*X0).c7+=x;
   (*X1).c7+=x;

   (*X0).c8=(*m0).u[35]-(*m0).u[17];
   (*X1).c8=-(*X0).c8;
   x=(*m1).u[35]-(*m1).u[17];
   (*X0).c8+=x;
   (*X1).c8+=x;

   (*X0).c9=(*m0).u[16]-(*m0).u[34];
   (*X1).c9=-(*X0).c9;
   x=(*m1).u[16]-(*m1).u[34];
   (*X0).c9+=x;
   (*X1).c9+=x;
}


static void vec2pauli(weyl_dble *r,weyl_dble *s,pauli_dble *m)
{
   su3_vector_dble *r1,*r2,*s1,*s2;

   r1=(su3_vector_dble*)(&((*r).c1));
   r2=(su3_vector_dble*)(&((*r).c2));
   s1=(su3_vector_dble*)(&((*s).c1));
   s2=(su3_vector_dble*)(&((*s).c2));
   
   (*m).u[ 0]=_re((*s1).c1,(*r1).c1)+_re((*s1).c1,(*r1).c1);
   (*m).u[ 1]=_re((*s1).c2,(*r1).c2)+_re((*s1).c2,(*r1).c2);
   (*m).u[ 2]=_re((*s1).c3,(*r1).c3)+_re((*s1).c3,(*r1).c3);

   (*m).u[ 3]=_re((*s2).c1,(*r2).c1)+_re((*s2).c1,(*r2).c1);
   (*m).u[ 4]=_re((*s2).c2,(*r2).c2)+_re((*s2).c2,(*r2).c2);
   (*m).u[ 5]=_re((*s2).c3,(*r2).c3)+_re((*s2).c3,(*r2).c3);
   
   (*m).u[ 6]=_re((*s1).c1,(*r1).c2)+_re((*r1).c1,(*s1).c2);
   (*m).u[ 7]=_im((*s1).c1,(*r1).c2)+_im((*r1).c1,(*s1).c2);
   (*m).u[ 8]=_re((*s1).c1,(*r1).c3)+_re((*r1).c1,(*s1).c3);
   (*m).u[ 9]=_im((*s1).c1,(*r1).c3)+_im((*r1).c1,(*s1).c3);      

   (*m).u[10]=_re((*s1).c1,(*r2).c1)+_re((*r1).c1,(*s2).c1);
   (*m).u[11]=_im((*s1).c1,(*r2).c1)+_im((*r1).c1,(*s2).c1);
   (*m).u[12]=_re((*s1).c1,(*r2).c2)+_re((*r1).c1,(*s2).c2);
   (*m).u[13]=_im((*s1).c1,(*r2).c2)+_im((*r1).c1,(*s2).c2);
   (*m).u[14]=_re((*s1).c1,(*r2).c3)+_re((*r1).c1,(*s2).c3);
   (*m).u[15]=_im((*s1).c1,(*r2).c3)+_im((*r1).c1,(*s2).c3);
   
   (*m).u[16]=_re((*s1).c2,(*r1).c3)+_re((*r1).c2,(*s1).c3);
   (*m).u[17]=_im((*s1).c2,(*r1).c3)+_im((*r1).c2,(*s1).c3);

   (*m).u[18]=_re((*s1).c2,(*r2).c1)+_re((*r1).c2,(*s2).c1);
   (*m).u[19]=_im((*s1).c2,(*r2).c1)+_im((*r1).c2,(*s2).c1);
   (*m).u[20]=_re((*s1).c2,(*r2).c2)+_re((*r1).c2,(*s2).c2);
   (*m).u[21]=_im((*s1).c2,(*r2).c2)+_im((*r1).c2,(*s2).c2);
   (*m).u[22]=_re((*s1).c2,(*r2).c3)+_re((*r1).c2,(*s2).c3);
   (*m).u[23]=_im((*s1).c2,(*r2).c3)+_im((*r1).c2,(*s2).c3);

   (*m).u[24]=_re((*s1).c3,(*r2).c1)+_re((*r1).c3,(*s2).c1);
   (*m).u[25]=_im((*s1).c3,(*r2).c1)+_im((*r1).c3,(*s2).c1);
   (*m).u[26]=_re((*s1).c3,(*r2).c2)+_re((*r1).c3,(*s2).c2);
   (*m).u[27]=_im((*s1).c3,(*r2).c2)+_im((*r1).c3,(*s2).c2);
   (*m).u[28]=_re((*s1).c3,(*r2).c3)+_re((*r1).c3,(*s2).c3);
   (*m).u[29]=_im((*s1).c3,(*r2).c3)+_im((*r1).c3,(*s2).c3);
   
   (*m).u[30]=_re((*s2).c1,(*r2).c2)+_re((*r2).c1,(*s2).c2);
   (*m).u[31]=_im((*s2).c1,(*r2).c2)+_im((*r2).c1,(*s2).c2);
   (*m).u[32]=_re((*s2).c1,(*r2).c3)+_re((*r2).c1,(*s2).c3);
   (*m).u[33]=_im((*s2).c1,(*r2).c3)+_im((*r2).c1,(*s2).c3);      
   (*m).u[34]=_re((*s2).c2,(*r2).c3)+_re((*r2).c2,(*s2).c3);
   (*m).u[35]=_im((*s2).c2,(*r2).c3)+_im((*r2).c2,(*s2).c3);
}


void vec2u3alg(spinor_dble *r,spinor_dble *s,u3_alg_dble *X)
{
   weyl_dble *v,*w;
   pauli_dble *m;

   m=ms;
   v=(weyl_dble*)(&((*r).c1));
   w=(weyl_dble*)(&((*s).c1));
   vec2pauli(v,w,m);

   m=ms+1;
   v=(weyl_dble*)(&((*r).c3));
   w=(weyl_dble*)(&((*s).c3));
   vec2pauli(v,w,m);

   det2u3alg(ms,X);
}


static void set2mat(su3_dble *u)
{
   (*u).c11.re=_re(psi1.c1,chi1.c1)+_re(psi2.c1,chi2.c1);
   (*u).c11.im=_im(psi1.c1,chi1.c1)+_im(psi2.c1,chi2.c1);    
   (*u).c12.re=_re(psi1.c1,chi1.c2)+_re(psi2.c1,chi2.c2);
   (*u).c12.im=_im(psi1.c1,chi1.c2)+_im(psi2.c1,chi2.c2);    
   (*u).c13.re=_re(psi1.c1,chi1.c3)+_re(psi2.c1,chi2.c3);
   (*u).c13.im=_im(psi1.c1,chi1.c3)+_im(psi2.c1,chi2.c3);    

   (*u).c21.re=_re(psi1.c2,chi1.c1)+_re(psi2.c2,chi2.c1);
   (*u).c21.im=_im(psi1.c2,chi1.c1)+_im(psi2.c2,chi2.c1);    
   (*u).c22.re=_re(psi1.c2,chi1.c2)+_re(psi2.c2,chi2.c2);
   (*u).c22.im=_im(psi1.c2,chi1.c2)+_im(psi2.c2,chi2.c2);    
   (*u).c23.re=_re(psi1.c2,chi1.c3)+_re(psi2.c2,chi2.c3);
   (*u).c23.im=_im(psi1.c2,chi1.c3)+_im(psi2.c2,chi2.c3);    
   
   (*u).c31.re=_re(psi1.c3,chi1.c1)+_re(psi2.c3,chi2.c1);
   (*u).c31.im=_im(psi1.c3,chi1.c1)+_im(psi2.c3,chi2.c1);    
   (*u).c32.re=_re(psi1.c3,chi1.c2)+_re(psi2.c3,chi2.c2);
   (*u).c32.im=_im(psi1.c3,chi1.c2)+_im(psi2.c3,chi2.c2);    
   (*u).c33.re=_re(psi1.c3,chi1.c3)+_re(psi2.c3,chi2.c3);
   (*u).c33.im=_im(psi1.c3,chi1.c3)+_im(psi2.c3,chi2.c3);    
}


static void add2mat(su3_dble *u)
{
   (*u).c11.re+=_re(psi1.c1,chi1.c1)+_re(psi2.c1,chi2.c1);
   (*u).c11.im+=_im(psi1.c1,chi1.c1)+_im(psi2.c1,chi2.c1);    
   (*u).c12.re+=_re(psi1.c1,chi1.c2)+_re(psi2.c1,chi2.c2);
   (*u).c12.im+=_im(psi1.c1,chi1.c2)+_im(psi2.c1,chi2.c2);    
   (*u).c13.re+=_re(psi1.c1,chi1.c3)+_re(psi2.c1,chi2.c3);
   (*u).c13.im+=_im(psi1.c1,chi1.c3)+_im(psi2.c1,chi2.c3);    

   (*u).c21.re+=_re(psi1.c2,chi1.c1)+_re(psi2.c2,chi2.c1);
   (*u).c21.im+=_im(psi1.c2,chi1.c1)+_im(psi2.c2,chi2.c1);    
   (*u).c22.re+=_re(psi1.c2,chi1.c2)+_re(psi2.c2,chi2.c2);
   (*u).c22.im+=_im(psi1.c2,chi1.c2)+_im(psi2.c2,chi2.c2);    
   (*u).c23.re+=_re(psi1.c2,chi1.c3)+_re(psi2.c2,chi2.c3);
   (*u).c23.im+=_im(psi1.c2,chi1.c3)+_im(psi2.c2,chi2.c3);    
   
   (*u).c31.re+=_re(psi1.c3,chi1.c1)+_re(psi2.c3,chi2.c1);
   (*u).c31.im+=_im(psi1.c3,chi1.c1)+_im(psi2.c3,chi2.c1);    
   (*u).c32.re+=_re(psi1.c3,chi1.c2)+_re(psi2.c3,chi2.c2);
   (*u).c32.im+=_im(psi1.c3,chi1.c2)+_im(psi2.c3,chi2.c2);    
   (*u).c33.re+=_re(psi1.c3,chi1.c3)+_re(psi2.c3,chi2.c3);
   (*u).c33.im+=_im(psi1.c3,chi1.c3)+_im(psi2.c3,chi2.c3);    
}


static void spin2mat0(spinor_dble *rx,spinor_dble *ry,
                      spinor_dble *sx,spinor_dble *sy,su3_dble *u)
{
   _vector_add(psi1,(*ry).c1,(*ry).c3);
   _vector_add(psi2,(*ry).c2,(*ry).c4);
   _vector_sub(chi1,(*sx).c1,(*sx).c3);
   _vector_sub(chi2,(*sx).c2,(*sx).c4);
   set2mat(u);

   _vector_add(psi1,(*sy).c1,(*sy).c3);
   _vector_add(psi2,(*sy).c2,(*sy).c4);
   _vector_sub(chi1,(*rx).c1,(*rx).c3);
   _vector_sub(chi2,(*rx).c2,(*rx).c4);
   add2mat(u);
}


static void spin2mat1(spinor_dble *rx,spinor_dble *ry,
                      spinor_dble *sx,spinor_dble *sy,su3_dble *u)
{
   _vector_i_add(psi1,(*ry).c1,(*ry).c4);
   _vector_i_add(psi2,(*ry).c2,(*ry).c3);
   _vector_i_sub(chi1,(*sx).c1,(*sx).c4);
   _vector_i_sub(chi2,(*sx).c2,(*sx).c3);
   set2mat(u);

   _vector_i_add(psi1,(*sy).c1,(*sy).c4);
   _vector_i_add(psi2,(*sy).c2,(*sy).c3);
   _vector_i_sub(chi1,(*rx).c1,(*rx).c4);
   _vector_i_sub(chi2,(*rx).c2,(*rx).c3);
   add2mat(u);
}

   
static void spin2mat2(spinor_dble *rx,spinor_dble *ry,
                      spinor_dble *sx,spinor_dble *sy,su3_dble *u)
{
   _vector_add(psi1,(*ry).c1,(*ry).c4);
   _vector_sub(psi2,(*ry).c2,(*ry).c3);
   _vector_sub(chi1,(*sx).c1,(*sx).c4);
   _vector_add(chi2,(*sx).c2,(*sx).c3);
   set2mat(u);

   _vector_add(psi1,(*sy).c1,(*sy).c4);
   _vector_sub(psi2,(*sy).c2,(*sy).c3);
   _vector_sub(chi1,(*rx).c1,(*rx).c4);
   _vector_add(chi2,(*rx).c2,(*rx).c3);
   add2mat(u);
}
   

static void spin2mat3(spinor_dble *rx,spinor_dble *ry,
                      spinor_dble *sx,spinor_dble *sy,su3_dble *u)
{
   _vector_i_add(psi1,(*ry).c1,(*ry).c3);
   _vector_i_sub(psi2,(*ry).c2,(*ry).c4);
   _vector_i_sub(chi1,(*sx).c1,(*sx).c3);
   _vector_i_add(chi2,(*sx).c2,(*sx).c4);
   set2mat(u);

   _vector_i_add(psi1,(*sy).c1,(*sy).c3);
   _vector_i_sub(psi2,(*sy).c2,(*sy).c4);
   _vector_i_sub(chi1,(*rx).c1,(*rx).c3);
   _vector_i_add(chi2,(*rx).c2,(*rx).c4);
   add2mat(u);
}


void (*spin2mat[4])(spinor_dble *rx,spinor_dble *ry,
                    spinor_dble *sx,spinor_dble *sy,su3_dble *u) =
{spin2mat0,spin2mat1,spin2mat2,spin2mat3};

