
/*******************************************************************************
*
* File flipbc.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs to switch the boundary conditions in the time direction
*
*   void flipbc(void)
*     Changes the sign of the time-like single-precision link variables at
*     time x0=L0*NPROC0-1
*
*   void flipbcd(void)
*     Changes the sign of the time-like double-precision link variables at
*     time x0=L0*NPROC0-1
*
* Notes:
*
* These programs act globally and should therefore be called simultaneously
* on all processes
*
* Only the basic field variables on the local lattices (those with address
* offsets less than 4*VOLUME) are flipped, while those in the boundary
* buffers are unchanged
*
* The programs for the Dirac operator on the *full* lattice will effectively
* implement its action for the case of anti-periodic boundary conditions in
* the time direction after flipping the time-like links
*
*******************************************************************************/

#define FLIPBC_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "global.h"

#define _su3_flip_sign(u) \
   (u).c11.re=-(u).c11.re; \
   (u).c11.im=-(u).c11.im; \
   (u).c12.re=-(u).c12.re; \
   (u).c12.im=-(u).c12.im; \
   (u).c13.re=-(u).c13.re; \
   (u).c13.im=-(u).c13.im; \
   (u).c21.re=-(u).c21.re; \
   (u).c21.im=-(u).c21.im; \
   (u).c22.re=-(u).c22.re; \
   (u).c22.im=-(u).c22.im; \
   (u).c23.re=-(u).c23.re; \
   (u).c23.im=-(u).c23.im; \
   (u).c31.re=-(u).c31.re; \
   (u).c31.im=-(u).c31.im; \
   (u).c32.re=-(u).c32.re; \
   (u).c32.im=-(u).c32.im; \
   (u).c33.re=-(u).c33.re; \
   (u).c33.im=-(u).c33.im

static int init=0,nlks=0,*lks=NULL;


static void alloc_lks(void)
{
   int ix,iy,*idx,*lk;

   if ((cpr[0]==0)||(cpr[0]==(NPROC0-1)))
   {
      if (NPROC0>1)
         nlks=(L1*L2*L3)/2;
      else
         nlks=L1*L2*L3;

      lks=amalloc(nlks*sizeof(int),3);
      idx=amalloc((VOLUME/2)*sizeof(int),3);

      if (error_loc((lks==NULL)||(idx==NULL),1,"alloc_lks [flipbc.c]",
                    "Unable to allocate index arrays")==0)
      {
         for (ix=0;ix<(VOLUME/2);ix++)
            idx[ix]=0;

         for (ix=0;ix<(L1*L2*L3);ix++)
         {
            if (cpr[0]==0)
            {
               iy=ipt[ix]-(VOLUME/2);

               if (iy>=0)
                  idx[iy]=-1;
            }

            if (cpr[0]==(NPROC0-1))
            {
               iy=ipt[ix+(L0-1)*L1*L2*L3]-(VOLUME/2);

               if (iy>=0)
                  idx[iy]=1;
            }
         }

         lk=lks;

         for (ix=0;ix<(VOLUME/2);ix++)
         {
            if (idx[ix]==-1)
            {
               *lk=8*ix+1;
               lk+=1;
            }

            if (idx[ix]==1)
            {
               *lk=8*ix;
               lk+=1;
            }
         }

         afree(idx);
      }
   }

   error_chk();
   init=1;
}


void flipbc(void)
{
   int *lk,*lkm;
   su3 *ub,*u;

   if (init==0)
      alloc_lks();

   ub=pu[VOLUME/2][0];
   error(ub==NULL,1,"flipbc [flipbc.c]",
         "Attempt to access unallocated memory space");
   lk=lks;
   lkm=lk+nlks;

   for (;lk<lkm;lk++)
   {
      u=ub+(*lk);
      _su3_flip_sign(*u);
   }

   set_flags(FLIPPED_BC);
}


void flipbcd(void)
{
   int *lk,*lkm;
   su3_dble *udb,*ud;

   if (init==0)
      alloc_lks();

   udb=pud[VOLUME/2][0];
   error(udb==NULL,1,"flipbcd [flipbc.c]",
         "Attempt to access unallocated memory space");
   lk=lks;
   lkm=lk+nlks;

   for (;lk<lkm;lk++)
   {
      ud=udb+(*lk);
      _su3_flip_sign(*ud);
   }

   set_flags(FLIPPED_BCD);
}
